"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .basesdk import BaseSDK
from .httpclient import AsyncHttpClient, HttpClient
from .sdkconfiguration import SDKConfiguration
from .utils.logger import Logger, get_default_logger
from .utils.retries import RetryConfig
from honeyhive import utils
from honeyhive._hooks import SDKHooks
from honeyhive.configurations import Configurations
from honeyhive.datapoints import Datapoints
from honeyhive.datasets import Datasets
from honeyhive.events import Events
from honeyhive.experiments import Experiments
from honeyhive.metrics import Metrics
from honeyhive.models import components
from honeyhive.projects import Projects
from honeyhive.session import Session
from honeyhive.tools import Tools
from honeyhive.types import OptionalNullable, UNSET
import httpx
from typing import Any, Callable, Dict, Optional, Union


class HoneyHive(BaseSDK):
    session: Session
    events: Events
    metrics: Metrics
    tools: Tools
    datapoints: Datapoints
    datasets: Datasets
    projects: Projects
    experiments: Experiments
    configurations: Configurations

    def __init__(
        self,
        bearer_auth: Union[str, Callable[[], str]],
        server_idx: Optional[int] = None,
        server_url: Optional[str] = None,
        url_params: Optional[Dict[str, str]] = None,
        client: Optional[HttpClient] = None,
        async_client: Optional[AsyncHttpClient] = None,
        retry_config: OptionalNullable[RetryConfig] = UNSET,
        timeout_ms: Optional[int] = None,
        debug_logger: Optional[Logger] = None,
    ) -> None:
        r"""Instantiates the SDK configuring it with the provided parameters.

        :param bearer_auth: The bearer_auth required for authentication
        :param server_idx: The index of the server to use for all methods
        :param server_url: The server URL to use for all methods
        :param url_params: Parameters to optionally template the server URL with
        :param client: The HTTP client to use for all synchronous methods
        :param async_client: The Async HTTP client to use for all asynchronous methods
        :param retry_config: The retry configuration to use for all supported methods
        :param timeout_ms: Optional request timeout applied to each operation in milliseconds
        """
        if client is None:
            client = httpx.Client()

        assert issubclass(
            type(client), HttpClient
        ), "The provided client must implement the HttpClient protocol."

        if async_client is None:
            async_client = httpx.AsyncClient()

        if debug_logger is None:
            debug_logger = get_default_logger()

        assert issubclass(
            type(async_client), AsyncHttpClient
        ), "The provided async_client must implement the AsyncHttpClient protocol."

        security: Any = None
        if callable(bearer_auth):
            security = lambda: components.Security(bearer_auth=bearer_auth())  # pylint: disable=unnecessary-lambda-assignment
        else:
            security = components.Security(bearer_auth=bearer_auth)

        if server_url is not None:
            if url_params is not None:
                server_url = utils.template_url(server_url, url_params)

        BaseSDK.__init__(
            self,
            SDKConfiguration(
                client=client,
                async_client=async_client,
                security=security,
                server_url=server_url,
                server_idx=server_idx,
                retry_config=retry_config,
                timeout_ms=timeout_ms,
                debug_logger=debug_logger,
            ),
        )

        hooks = SDKHooks()

        current_server_url, *_ = self.sdk_configuration.get_server_details()
        server_url, self.sdk_configuration.client = hooks.sdk_init(
            current_server_url, self.sdk_configuration.client
        )
        if current_server_url != server_url:
            self.sdk_configuration.server_url = server_url

        # pylint: disable=protected-access
        self.sdk_configuration.__dict__["_hooks"] = hooks

        self._init_sdks()

    def _init_sdks(self):
        self.session = Session(self.sdk_configuration)
        self.events = Events(self.sdk_configuration)
        self.metrics = Metrics(self.sdk_configuration)
        self.tools = Tools(self.sdk_configuration)
        self.datapoints = Datapoints(self.sdk_configuration)
        self.datasets = Datasets(self.sdk_configuration)
        self.projects = Projects(self.sdk_configuration)
        self.experiments = Experiments(self.sdk_configuration)
        self.configurations = Configurations(self.sdk_configuration)




from honeyhive.tracer.custom import trace, atrace, enrich_span

import asyncio
from dataclasses import dataclass, field, fields
from typing import Any, Optional, Callable, Coroutine
import inspect
import functools
import concurrent.futures
import json

class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

# ------------------------------------------------------------------------------
# EVALUATOR SETTINGS
# ------------------------------------------------------------------------------



EVALUATOR_SETTINGS_KEYS = [
    'wraps', 
    'weight', 
    'asserts', 
    'repeat', 
    'transform', 
    'aggregate', 
    'checker', 
    'target', 
    'evaluate'
]


@dataclass
class EvalSettings:
    name: str  # config key
    wraps: Optional[str | dict] = None
    weight: float = None
    asserts: bool = None
    repeat: Optional[int] = None
    transform: Optional[str] = None
    aggregate: Optional[str] = None
    checker: Optional[str] = None
    target: Optional[str] = None
    evaluate: Optional[str] = None

    def copy(self) -> "EvalSettings":
        return EvalSettings(
            name=self.name,
            wraps=self.wraps,
            weight=self.weight,
            repeat=self.repeat,
            asserts=self.asserts,
            transform=self.transform,
            aggregate=self.aggregate,
            checker=self.checker,
            target=self.target,
            evaluate=self.evaluate,
        )
    
    def keys(self):
        return self.__dict__.keys()

    # TODO: settings update should replace instead of merge
    def update(self, eval_settings: Any | None) -> None:
        if eval_settings is None:
            return
        if isinstance(eval_settings, dict):
            update_dict = eval_settings
        elif isinstance(eval_settings, EvalSettings):
            update_dict = eval_settings.__dict__
        else:
            raise TypeError(
                "eval_settings must be either a dictionary or an EvalSettings instance. Got {}".format(
                    type(eval_settings)
                )
            )

        valid_fields = {f.name for f in fields(self)}

        for key, value in update_dict.items():
            if key not in valid_fields:
                raise ValueError(f"Invalid field name: {key}")
            if value is not None:
                setattr(self, key, value)
    
    @staticmethod
    def extract_eval_settings_and_kwargs(settings: dict[str, Any] | None):

        eval_kwargs = dict()
        eval_settings = dict()

        if settings is not None:
            for key, value in settings.items():
                if key in EVALUATOR_SETTINGS_KEYS:
                    eval_settings[key] = value
                else:
                    eval_kwargs[key] = value

        return eval_settings, eval_kwargs

    def __str__(self) -> str:
        dict_str = {k: str(v) for k, v in self.__dict__.items() if v is not None}
        return json.dumps(dict_str, indent=4).replace('"', "")

    def dict(self) -> dict:
        ret_dict = self.__dict__
        ret_dict.pop('name', None)
        return ret_dict

@dataclass
class EvaluatorSettings:
    name: str
    
    # base default settings
    default_settings: EvalSettings = None
    default_kwargs: dict = field(default_factory=dict)
    
    # settings from defaults.yaml
    defaults_yaml_settings: EvalSettings = None
    defaults_yaml_kwargs: dict = field(default_factory=dict)
    
    # settings from evaluator init
    init_settings: EvalSettings = None
    init_kwargs: dict = field(default_factory=dict)
    
    # settings from decorator kwargs
    deco_settings: EvalSettings = None
    deco_kwargs: dict = field(default_factory=dict)
    
    # settings from config file
    config_settings: EvalSettings = None
    config_kwargs: dict = field(default_factory=dict)
    
    # settings from runtime
    explicit_settings: EvalSettings = None
    explicit_kwargs: dict = field(default_factory=dict)
    

    def __post_init__(self):
        self.default_settings = EvalSettings(name=self.name)
        self.defaults_yaml_settings = EvalSettings(name=self.name)
        self.init_settings = EvalSettings(name=self.name)
        self.deco_settings = EvalSettings(name=self.name)
        self.config_settings = EvalSettings(name=self.name)
        self.explicit_settings = None # this must be set at runtime
        
        # set defaults
        self.default_settings.asserts = False
        self.default_settings.weight = 1.0

    def resolve_settings(self, settings: EvalSettings | None = None) -> EvalSettings:
        if self.explicit_settings:
            return self.explicit_settings
        
        final_settings: EvalSettings = self.default_settings
        final_settings.update(self.defaults_yaml_settings)
        final_settings.update(self.init_settings)
        final_settings.update(self.config_settings)
        final_settings.update(self.deco_settings)
        
        if settings is not None:
            final_settings.update(settings)
        
        return final_settings
    
    def resolve_kwargs(self, kwargs: dict | None = None) -> dict:
        if self.explicit_kwargs:
            return self.explicit_kwargs
        
        final_kwargs: dict = self.default_kwargs.copy()
        final_kwargs.update(self.defaults_yaml_kwargs)
        final_kwargs.update(self.init_kwargs)
        final_kwargs.update(self.config_kwargs)
        final_kwargs.update(self.deco_kwargs)
        
        if kwargs is not None:
            final_kwargs.update(kwargs)

        return final_kwargs
        
# ------------------------------------------------------------------------------
# EVALUATOR RESULT
# ------------------------------------------------------------------------------


class EvalResult:

    def __init__(self, score: Any, init_method: Optional[str] = None, **metadata):

        self.score: Any | EvalResult = score
        self.metadata: dict = metadata

        # determine the eval_type
        self.init_method = init_method or inspect.stack()[1].function

        self.eval_settings: Optional[EvalSettings] = EvalSettings(name=self.init_method)
        self.eval_kwargs: Optional[dict] = dict()

        # for easy access
        self.weight = self.eval_settings.weight

        self.func_impl: Callable = None
        self.func_args: tuple = None
        self.func_kwargs: dict = None

    def to_dict(self) -> dict:
        return {"score": self.score, "metadata": self.metadata}

    def copy(self) -> "EvalResult":
        copy_result = EvalResult(score=self.score,
                          init_method=self.init_method,
                          **self.metadata)
        return copy_result

class EvaluatorMeta(type):
    
    def __getattribute__(cls, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return cls.__class_getitem__(name)
        

class evaluator(metaclass=EvaluatorMeta):

    # ------------------------------------------------------------------------------
    # STATICS / INITIALIZE
    # ------------------------------------------------------------------------------

    # global registry of evaluator names to evaluator instances
    all_evaluators: dict[str, "evaluator" | Callable | Coroutine | "aevaluator"] = dict()

    # global registry of evaluator names to evaluator settings
    all_evaluator_settings: dict[str, EvaluatorSettings] = dict()

    def __unnamed__(self, *args, **kwargs):
        raise NotImplementedError(f"Please decorate with an evaluator implementation.")

    def __new__(
        cls, 
        func=None, 
        eval_settings=None, 
        eval_kwargs=None, 
        **deco_settings_kwargs
    ) -> "evaluator":
        """Allows evaluator to be initialized in the decorator with kwargs"""
        if func is None:
            return lambda f: cls(f, eval_settings, eval_kwargs, **deco_settings_kwargs)
        return super().__new__(cls)

    def __init__(
        self,
        func: Callable = __unnamed__,
        eval_settings: EvalSettings | None = None,
        eval_kwargs: dict[str, Any] | None = None,
        **deco_settings_kwargs,
    ) -> None:
        
        # set the wrapped function implementation and its name
        self.func: Callable = func

        # set the evaluator name
        self.name: str = func.__name__ if hasattr(func, "__name__") else str(func)
        
        # register all_evaluators[func name] = this evaluator
        self.all_evaluators[self.name] = self
        
        # initialize the evaluator settings
        if self.name not in evaluator.all_evaluator_settings:
            evaluator.all_evaluator_settings[self.name] = EvaluatorSettings(self.name)

        # get the settings
        evaluator_settings = evaluator.all_evaluator_settings[self.name]
            
        # init settings
        if eval_settings:
            eval_settings.name = self.name
        else:
            eval_settings = EvalSettings(name=self.name)
        
        evaluator_settings.init_settings = eval_settings
        evaluator_settings.init_kwargs = eval_kwargs or {}
        
        # decorator kwargs
        kwarg_eval_settings, kwarg_eval_kwargs = EvalSettings.extract_eval_settings_and_kwargs(
            deco_settings_kwargs
        )
        kwarg_eval_settings['name'] = self.name
        evaluator_settings.deco_settings = EvalSettings(**kwarg_eval_settings)
        evaluator_settings.deco_kwargs = kwarg_eval_kwargs
        
        self.explicit_config = None

        # sets decorator metadata to that of wrapped function
        functools.update_wrapper(self, func)
        functools.update_wrapper(func, self)


    # ------------------------------------------------------------------------------
    # AGGREGATION
    def pre_apply_aggregation(
        self,
        eval_results: tuple[EvalResult] | list[EvalResult],
        eval_scores: tuple | list,
        final_settings: EvalSettings,
    ) -> tuple[EvalResult, Any] | Coroutine:

        locals_dict = {
            "values": eval_scores, 
            "results": eval_results
        }

        # TODO: enable aggregate to be a function
        aggregation_expr = str(final_settings.aggregate)

        # apply aggregation
        aggregate_score = eval(
            aggregation_expr, 
            evaluator.all_evaluators, 
            locals_dict
        )

        return aggregate_score

    def post_apply_aggregation(
        self,
        eval_results: tuple[EvalResult] | list[EvalResult] | EvalResult,
        aggregate_score: Any,
        final_settings: EvalSettings,
    ):
        init_methods = set()

        # if no repetitions, we will only have one eval result
        if isinstance(eval_results, EvalResult):
            init_methods.add(eval_results.init_method)
        else:
            for eval_result in eval_results:
                if isinstance(eval_result, EvalResult):
                    init_methods.add(eval_result.init_method)

        init_method = "aggregate: "
        if len(init_methods) > 0:
            init_method += "-".join(init_methods)

        aggregate_result = EvalResult(
            aggregate_score, 
            init_method=init_method, 
            prev_results=eval_results
        )

        return aggregate_result

    def sync_apply_aggregation(
        self,
        eval_results: tuple[EvalResult] | list[EvalResult] | EvalResult,
        eval_scores: tuple | list | Any,
        final_settings: EvalSettings,
    ) -> tuple[EvalResult, Any]:

        if not final_settings.aggregate:
            return eval_results, eval_scores

        aggregate_score = self.pre_apply_aggregation(
            eval_results, 
            eval_scores, 
            final_settings
        )
        
        aggregate_score = evaluator.resolve_pipeline(aggregate_score, eval_scores)
            
        aggregate_result = self.post_apply_aggregation(
            eval_results, 
            aggregate_score, 
            final_settings
        )

        return aggregate_result, aggregate_score

    async def async_apply_aggregation(
        self,
        eval_results: tuple[EvalResult] | list[EvalResult],
        eval_scores: tuple | list,
        final_settings: EvalSettings,
    ) -> tuple[EvalResult, Any]:

        if not final_settings.aggregate:
            return eval_results, eval_scores

        aggregate_score = self.pre_apply_aggregation(
            eval_results, 
            eval_scores, 
            final_settings
        )

        aggregate_score = await evaluator.aresolve_pipeline(aggregate_score, eval_scores)

        aggregate_result = self.post_apply_aggregation(
            eval_results, 
            aggregate_score, 
            final_settings
        )

        return aggregate_result, aggregate_score
    # ------------------------------------------------------------------------------


    # ------------------------------------------------------------------------------
    # TRANSFORMATION
    def pre_apply_transformation(
        self, 
        eval_result: EvalResult, 
        eval_score: Any,
        final_settings: EvalSettings,
    ):

        transform_expr = str(final_settings.transform)

        locals_dict = {
            "value": eval_score, 
            "result": eval_result
        }

        # apply transformation
        transformed_score = eval(
            transform_expr, 
            evaluator.all_evaluators, 
            locals_dict
        )

        return transformed_score

    def post_apply_transformation(
        self, 
        eval_result: EvalResult, 
        transformed_score: Any,
        final_settings: EvalSettings,
    ):
        init_method = "transform: " + eval_result.init_method

        transformed_result = EvalResult(
            transformed_score, 
            init_method=init_method, 
            prev_result=eval_result
        )
        
        transformed_result.weight = final_settings.weight

        return transformed_result

    async def async_apply_transformation(
        self,
        eval_result: EvalResult,
        eval_score: Any,
        final_settings: EvalSettings,
    ) -> tuple[EvalResult, Any]:

        if not final_settings.transform:
            return eval_result, eval_score

        transformed_score = self.pre_apply_transformation(
            eval_result, 
            eval_score, 
            final_settings
        )

        transformed_score = await evaluator.aresolve_pipeline(transformed_score, eval_score)

        transformed_result = self.post_apply_transformation(
            eval_result, 
            transformed_score, 
            final_settings
        )

        return transformed_result, transformed_score

    def sync_apply_transformation(
        self,
        eval_result: EvalResult,
        eval_score: Any,
        final_settings: EvalSettings,
    ) -> tuple[EvalResult, Any]:

        if not final_settings.transform:
            return eval_result, eval_score

        transformed_score = self.pre_apply_transformation(
            eval_result, eval_score, final_settings
        )
        
        transformed_score = evaluator.resolve_pipeline(transformed_score, eval_score)

        transformed_result = self.post_apply_transformation(
            eval_result, transformed_score, final_settings
        )

        return transformed_result, transformed_score
    # ------------------------------------------------------------------------------


    # ------------------------------------------------------------------------------
    # CHECKER
    def pre_run_checker(
        self,
        eval_result: EvalResult,
        eval_score: Any,
        final_settings: EvalSettings,
    ) -> bool:

        checker_expr = str(final_settings.checker)

        locals_dict = {
            "value": eval_score, 
            "result": eval_result, 
            "target": final_settings.target
        }

        # evaluate checker
        checker_score = eval(
            checker_expr, 
            evaluator.all_evaluators, 
            locals_dict
        )

        return checker_score

    def post_run_checker(
        self,
        eval_result: EvalResult,
        eval_score: Any,
        final_settings: EvalSettings,
        checker_score: Any = None,
    ) -> bool:

        if final_settings.asserts:
            assert (
                checker_score
            ), f"Assertion failed: score {eval_score} is not in range {final_settings.target}"

        init_method = "checker: " + eval_result.init_method

        checker_result = EvalResult(
            checker_score, 
            init_method=init_method, 
            prev_result=eval_result
        )

        return checker_result

    def run_asserts(
        self,
        eval_score: Any,
        final_settings: EvalSettings,
    ) -> bool:
        
        if final_settings.target is None:
            failure_message = f"Assertion failed: score {eval_score}"
        else:
            failure_message = f"Assertion failed: score {eval_score} is not in range {final_settings.target}"
        
        assert eval_score, failure_message

    def run_checker(
        self,
        eval_result: EvalResult,
        eval_score: Any,
        final_settings: EvalSettings,
    ) -> bool:

        if not final_settings.checker:
            if not final_settings.asserts:
                return eval_result, eval_score

            self.run_asserts(eval_score, final_settings)

            return eval_result, eval_score

        checker_score = self.pre_run_checker(
            eval_result, 
            eval_score, 
            final_settings
        )
        
        checker_score = evaluator.resolve_pipeline(checker_score, eval_score, final_settings.target)

        checker_result = self.post_run_checker(
            eval_result, 
            eval_score, 
            final_settings, 
            checker_score
        )

        return checker_result, checker_score

    async def arun_checker(
        self,
        eval_result: EvalResult,
        eval_score: Any,
        final_settings: EvalSettings,
    ) -> bool:

        if not final_settings.checker:
            if not final_settings.asserts:
                return eval_result, eval_score

            self.run_asserts(eval_score, final_settings)

            return eval_result, eval_score

        checker_score = self.pre_run_checker(
            eval_result, 
            eval_score,
            final_settings
        )

        checker_score = await evaluator.aresolve_pipeline(
            checker_score, 
            eval_score, 
            final_settings.target
        )

        checker_result = self.post_run_checker(
            eval_result, 
            eval_score, 
            final_settings, 
            checker_score
        )

        return checker_result, checker_score
    # ------------------------------------------------------------------------------
    
    
    # ------------------------------------------------------------------------------
    # WRAPPING
    @staticmethod
    def parse_wraps(wraps: str | dict | None | Any):
        if wraps is None:
            return None, {}
        
        if isinstance(wraps, str):
            return wraps, {}
        elif isinstance(wraps, dict):
            # assert that there is a single key of type str
            assert len(wraps) == 1 and isinstance(list(wraps.keys())[0], str), \
                "wraps must be a single key of type str"
            
            wrapped_eval_name = list(wraps.keys())[0]
            wrapped_eval_settings_kwargs = wraps[wrapped_eval_name]
            return wrapped_eval_name, wrapped_eval_settings_kwargs
        else:
            raise ValueError(f"Invalid wraps type: {type(wraps)}")
    
    @staticmethod
    def create_wrapper(
        base_callable: 'evaluator',
        wrapped_eval_settings: EvalSettings,
        wrapped_eval_kwargs: dict,
        wrapper_name: str,
    ) -> Callable:
        """
        Create a wrapper function for an evaluator, given the base evaluator,
        the wrapped evaluator settings, the wrapped evaluator kwargs, and the wrapper name.
        
        The wrapped_eval / base_callable's settings and kwargs update any previous settings and kwargs.
        The wrapper's settings do NOT update the wrapped evaluator's settings.
        The wrapper's kwargs DO update the wrapped evaluator's kwargs.
        
        The final settings and kwargs are passed into the wrapped evaluator during calltime.
        Due to the ordering of the dict unpacking, the wrapper's kwargs will update the 
        wrapped evaluator's kwargs. The settings are also passed as kwargs into the base callable.

        Args:
            base_callable (evaluator): The base evaluator to be wrapped.
            wrapped_eval_settings (EvalSettings): Settings for the wrapped evaluator.
            wrapped_eval_kwargs (dict): Additional keyword arguments for the wrapped evaluator.
            wrapper_name (str): Name for the wrapper function.

        Returns:
            Callable: A wrapper function that calls the base evaluator with the provided settings and arguments.
        """
        
        
        base_callable_settings = wrapped_eval_settings.copy()
        base_callable_kwargs = wrapped_eval_kwargs.copy()
        
        if asyncio.iscoroutinefunction(base_callable.func):
            async def afunc(*args, **kwargs):
                return await base_callable(
                    *args,
                    ** {
                        **base_callable_settings.dict(),
                        **base_callable_kwargs,
                        **kwargs
                    }
                )
            afunc.__name__ = wrapper_name
            return afunc
        
        def func(*args, **kwargs):
            return base_callable(
                *args,
                ** {
                    **base_callable_settings.dict(),
                    **base_callable_kwargs,
                    **kwargs
                }
            )
        func.__name__ = wrapper_name
        return func
    
    @staticmethod
    def create_wrapped_evaluator(
        evaluator_settings: EvaluatorSettings
    ) -> None:
        
        wrapper_name = evaluator_settings.name
        wrapper_settings = evaluator_settings.resolve_settings()
        wrapper_kwargs = evaluator_settings.resolve_kwargs()
        
        # parse the wrapped evaluator
        wrapped_eval_name, wrapped_eval_settings_kwargs = evaluator.parse_wraps(
            wrapper_settings.wraps
        )
        assert isinstance(wrapped_eval_name, str), \
            f"wrapped evaluator name must be a string but got: {type(wrapped_eval_name)}"
        
        wrapped_eval_settings, wrapped_eval_kwargs = \
            EvalSettings.extract_eval_settings_and_kwargs(wrapped_eval_settings_kwargs)
        
        # create the wrapped evaluator settings if not already created
        # this can happen if the wrapped evaluator is not defined in any config file
        if wrapped_eval_name not in evaluator.all_evaluator_settings:
            evaluator.all_evaluator_settings[wrapped_eval_name] = EvaluatorSettings(name=wrapped_eval_name)
        
        
        # get the wrapped evaluator. It must be a registered evaluator
        base_callable = eval(
            wrapped_eval_name, 
            evaluator.all_evaluators,
        )
        
        # TODO: if the base callable is not a Callable but
        # not an evaluator, we need to wrap it as well
        assert isinstance(base_callable, evaluator)
        
        
        
        # update the wrapped evaluator settings with the wrapper settings
        final_wrapped_eval_settings = evaluator.all_evaluator_settings[wrapped_eval_name].resolve_settings().copy()
        final_wrapped_eval_settings.update(wrapped_eval_settings)
        
        # update the wrapped evaluator kwargs with the wrapper kwargs
        final_wrapped_eval_kwargs = evaluator.all_evaluator_settings[wrapped_eval_name].resolve_kwargs().copy()
        final_wrapped_eval_kwargs.update(wrapped_eval_kwargs)
        
        # the wrapper's kwargs merge with the wrapped evaluator's kwargs,
        # but the settings don't.
        final_wrapped_eval_kwargs.update(wrapper_kwargs)
        
        
        if asyncio.iscoroutinefunction(base_callable.func):
            afunc: Coroutine = evaluator.create_wrapper(
                base_callable,
                final_wrapped_eval_settings,
                final_wrapped_eval_kwargs,
                wrapper_name,
            )
            
            # initialize and register the wrapper eval
            aevaluator(func=afunc)
            
        else:
            # make the wrapper eval
            func: Callable = evaluator.create_wrapper(
                base_callable,
                final_wrapped_eval_settings,
                final_wrapped_eval_kwargs,
                wrapper_name,
            )

            # initialize and register the wrapper eval
            evaluator(func=func)
    # ------------------------------------------------------------------------------


    # ------------------------------------------------------------------------------
    # CALLING
    @staticmethod
    async def aresolve_pipeline(score, *args, **kwargs):
        if asyncio.iscoroutinefunction(score) or isinstance(score, aevaluator):
            score = await score(*args, **kwargs)
        elif isinstance(score, Callable):
            score = score(*args, **kwargs)

        return score
    
    @staticmethod
    def resolve_pipeline(score, *args, **kwargs):
        # string evaluated to a function
        if isinstance(score, Callable) or isinstance(score, evaluator):
            score = score(*args, **kwargs)
        return score
    
    def get_final_settings_and_kwargs(self, call_kwargs):
        eval_settings, eval_kwargs = EvalSettings.extract_eval_settings_and_kwargs(call_kwargs)
        explicit_settings, explicit_kwargs = EvalSettings.extract_eval_settings_and_kwargs(self.explicit_config)
        
        # calltime copy
        final_settings = self.settings.copy()
        final_settings.update(eval_settings)
        final_settings.update(explicit_settings)
        
        final_kwargs = self.kwargs.copy()
        final_kwargs.update(eval_kwargs)
        final_kwargs.update(explicit_kwargs)
        
        return final_settings, final_kwargs

    @atrace(event_type='chain', event_name='Evaluation')
    async def async_call(self, *call_args, **call_kwargs):
        
        final_settings, final_kwargs = self.get_final_settings_and_kwargs(call_kwargs)

        async def asingle_evaluation() -> tuple[EvalResult, Any]:

            # run the evaluator
            score = await atrace(self.func)(*call_args, **final_kwargs)

            result = EvalResult(
                score=score, 
                init_method=self.name,
                eval_settings=final_settings,
                eval_kwargs=final_kwargs,
                func_impl=f"{self.func.__module__}.{self.func.__name__}",
                func_args=call_args,
                func_kwargs=call_kwargs,
            )

            enrich_span(
                inputs={
                    "score": score,
                },
                outputs={
                    "result": result,
                },
                config={
                    "final_settings": str(final_settings),
                    "final_kwargs": final_kwargs,
                },
                metadata={
                    "func": self.name,
                    "func_impl": f"{self.func.__module__}.{self.func.__name__}",
                    "func_args": call_args,
                    "func_kwargs": call_kwargs,
                }
            )

            # transform
            transformed_result, transformed_score = (
                await self.async_apply_transformation(
                    result, 
                    score, 
                    final_settings
                )
            )

            # check target on transform if aggregate not defined
            if not final_settings.aggregate:
                checker_result, checker_score = await self.arun_checker(
                    eval_result=transformed_result,
                    eval_score=transformed_score,
                    final_settings=final_settings,
                )

                return checker_result, checker_score

            return transformed_result, transformed_score

        # execute repetition
        if final_settings.repeat:
            # Parallel evaluation
            results_scores = await asyncio.gather(
                *(
                    asingle_evaluation() 
                    for _ in range(final_settings.repeat)
                )
            )
            results, scores = zip(*results_scores)
            results = tuple(results)
            scores = tuple(scores)
        else:
            results, scores = await asingle_evaluation()

        # apply aggregation
        aggregate_result, aggregate_score = await self.async_apply_aggregation(
            results, 
            scores, 
            final_settings
        )

        # check target on aggregate if aggregate defined
        if final_settings.aggregate:
            checker_result, checker_score = await self.arun_checker(
                eval_result=aggregate_result,
                eval_score=aggregate_score,
                final_settings=final_settings,
            )

            return checker_result, checker_score

        return aggregate_result, aggregate_score

    @trace(event_type='chain', event_name='Evaluation')
    def sync_call(self, *call_args, **call_kwargs):
        
        final_settings, final_kwargs = self.get_final_settings_and_kwargs(call_kwargs)

        def single_evaluation() -> tuple[EvalResult, Any]:

            # run the evaluator
            score = self.func(*call_args, **final_kwargs)
            
            result = EvalResult(
                score=score, 
                init_method=self.name,
                eval_settings=final_settings,
                eval_kwargs=final_kwargs,
                func_impl=f"{self.func.__module__}.{self.func.__name__}",
                func_args=call_args,
                func_kwargs=call_kwargs,
            )

            enrich_span(
                inputs={
                    "score": score,
                },
                outputs={
                    "result": result,
                },
                config={
                    "final_settings": str(final_settings),
                    "final_kwargs": final_kwargs,
                },
                metadata={
                    "func": self.name,
                    "func_impl": f"{self.func.__module__}.{self.func.__name__}",
                    "func_args": call_args,
                    "func_kwargs": call_kwargs,
                }
            )

            # transform
            transformed_result, transformed_score = self.sync_apply_transformation(
                result, 
                score, 
                final_settings
            )

            # check target on transform if aggregate not defined
            if not final_settings.aggregate:
                checker_result, checker_score = self.run_checker(
                    eval_result=transformed_result,
                    eval_score=transformed_score,
                    final_settings=final_settings,
                )

                return checker_result, checker_score

            return transformed_result, transformed_score

        # execute repetition
        # TODO: add option for sequential evaluation since thread pools may not work for asyncio
        if final_settings.repeat:
            # Serial evaluation
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = [
                    executor.submit(single_evaluation)
                    for _ in range(final_settings.repeat)
                ]
                results, scores = zip(*[
                    future.result() 
                    for future in futures
                ])
            results = tuple(results)
            scores = tuple(scores)
        else:
            results, scores = single_evaluation()

        # apply aggregation
        aggregate_result, aggregate_score = self.sync_apply_aggregation(
            results, scores, final_settings
        )

        # check target on aggregate if aggregate defined
        if final_settings.aggregate:
            checker_result, checker_score = self.run_checker(
                eval_result=aggregate_result,
                eval_score=aggregate_score,
                final_settings=final_settings,
            )

            return checker_result, checker_score

        return aggregate_result, aggregate_score

    def __call__(self, *args, **kwargs) -> Any:
        
        # RUN EVALUATOR
        results, scores = None, None
        assert not asyncio.iscoroutinefunction(
            self.func
        ), "please use @aevaluator instead of @evaluator for this function"
        results, scores = self.sync_call(*args, **kwargs)

        return scores

    async def __acall__(self, *args, **kwargs) -> Any:
        
        # RUN EVALUATOR
        results, scores = None, None
        if asyncio.iscoroutinefunction(self.func):
            results, scores = await self.async_call(*args, **kwargs)
        else:
            results, scores = self.sync_call(*args, **kwargs)
        return scores    

    def raw(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    async def araw(self, *args, **kwargs):
        return await self.func(*args, **kwargs)
    
    # ------------------------------------------------------------------------------
    # PROPERTIES
    # ------------------------------------------------------------------------------

    @property
    def settings(self) -> EvalSettings:
        if self.name not in evaluator.all_evaluator_settings:
            evaluator.all_evaluator_settings[self.name] = EvaluatorSettings(name=self.name)
        return evaluator.all_evaluator_settings[self.name].resolve_settings()

    @settings.setter
    def settings(self, value: EvalSettings):
        assert isinstance(value, EvalSettings)
        if self.name not in evaluator.all_evaluator_settings:
            evaluator.all_evaluator_settings[self.name] = EvaluatorSettings(name=self.name)
        evaluator.all_evaluator_settings[self.name].explicit_settings = value

    @property
    def kwargs(self) -> dict[str, Any]:
        if self.name not in evaluator.all_evaluator_settings:
            evaluator.all_evaluator_settings[self.name] = EvaluatorSettings(name=self.name)
        return evaluator.all_evaluator_settings[self.name].resolve_kwargs()

    @kwargs.setter
    def kwargs(self, value: dict):
        assert isinstance(value, dict)
        if self.name not in evaluator.all_evaluator_settings:
            evaluator.all_evaluator_settings[self.name] = EvaluatorSettings(name=self.name)
        evaluator.all_evaluator_settings[self.name].explicit_kwargs = value

    @property
    def config(self) -> dict[str, Any]:
        if self.explicit_config is None:
            self.explicit_config = {
                **self.settings.dict(),
                **self.kwargs
            }
        return self.explicit_config
    
    @config.setter
    def config(self, value: Any):
        if value is None:
            self.explicit_config = None
        else:
            raise NotImplementedError

    
    # ------------------------------------------------------------------------------
    # ACCESSORS
    # ------------------------------------------------------------------------------

    @classmethod
    def _validate_key(cls, key: str | Callable | None) -> str:
        if isinstance(key, str):
            return key
        elif isinstance(key, Callable):
            if hasattr(key, "__name__"):
                return key.__name__
            else:
                return str(key)
        else:
            raise KeyError(f"Invalid key type: {type(key)}")

    @classmethod
    def __class_getitem__(cls, keys):
        '''
        Get an evaluator by name or Callable.
        '''
        
        if isinstance(keys, (str, Callable)):
            # Ensure that key is a string
            key: str = cls._validate_key(keys)
            if key in cls.all_evaluators:
                return cls.all_evaluators[key]
            elif key in cls.all_evaluator_settings:
                # if the evaluator is wrapped, initialize and register the wrapper
                if (evaluator_settings := cls.all_evaluator_settings[key]) is not None:
                    # initialize and register the wrapper
                    evaluator.create_wrapped_evaluator(evaluator_settings)
                    return cls.all_evaluators[key]
            else:
                raise KeyError(f"Key '{key}' not found in evaluators or config.")
        elif isinstance(keys, tuple):
            # Multiple key access
            return [cls.__class_getitem__(key) for key in keys]
        else:
            raise KeyError(f"Invalid key type: {type(keys)}")

    @classmethod
    def __class_setitem__(cls, key, value):
        key = cls._validate_key(key)
        cls.all_evaluators[key] = value

    @classmethod
    def __class_delitem__(cls, key):
        key = cls._validate_key(key)
        del cls.all_evaluators[key]

    @property
    def __code__(self):
        return self.func.__code__

class aevaluator(evaluator):

    async def __call__(self, *args, **kwargs):
        return await self.__acall__(*args, **kwargs)

    async def raw(self, *args, **kwargs):
        return await self.araw(*args, **kwargs)


# ------------------------------------------------------------------------------
# EVALUATORS
# ------------------------------------------------------------------------------

@evaluator
def mean(scores: list[float]) -> float:
    return sum(scores) / len(scores)

@evaluator
def median(scores: list[float]) -> float:
    return sorted(scores)[len(scores) // 2]

@evaluator
def mode(scores: list[float]) -> float:
    return max(set(scores), key=scores.count)




"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from enum import Enum
from honeyhive.types import BaseModel
from typing import Any, Dict, List, Optional, TypedDict
from typing_extensions import NotRequired


class CreateDatasetRequestType(str, Enum):
    r"""What the dataset is to be used for - \"evaluation\" (default) or \"fine-tuning\" """

    EVALUATION = "evaluation"
    FINE_TUNING = "fine-tuning"


class CreateDatasetRequestPipelineType(str, Enum):
    r"""The type of data included in the dataset - \"event\" (default) or \"session\" """

    EVENT = "event"
    SESSION = "session"


class CreateDatasetRequestTypedDict(TypedDict):
    project: str
    r"""Name of the project associated with this dataset like `New Project`"""
    name: str
    r"""Name of the dataset"""
    description: NotRequired[str]
    r"""A description for the dataset"""
    type: NotRequired[CreateDatasetRequestType]
    r"""What the dataset is to be used for - \"evaluation\" (default) or \"fine-tuning\" """
    datapoints: NotRequired[List[str]]
    r"""List of unique datapoint ids to be included in this dataset"""
    linked_evals: NotRequired[List[str]]
    r"""List of unique evaluation run ids to be associated with this dataset"""
    saved: NotRequired[bool]
    pipeline_type: NotRequired[CreateDatasetRequestPipelineType]
    r"""The type of data included in the dataset - \"event\" (default) or \"session\" """
    metadata: NotRequired[Dict[str, Any]]
    r"""Any helpful metadata to track for the dataset"""


class CreateDatasetRequest(BaseModel):
    project: str
    r"""Name of the project associated with this dataset like `New Project`"""

    name: str
    r"""Name of the dataset"""

    description: Optional[str] = None
    r"""A description for the dataset"""

    type: Optional[CreateDatasetRequestType] = None
    r"""What the dataset is to be used for - \"evaluation\" (default) or \"fine-tuning\" """

    datapoints: Optional[List[str]] = None
    r"""List of unique datapoint ids to be included in this dataset"""

    linked_evals: Optional[List[str]] = None
    r"""List of unique evaluation run ids to be associated with this dataset"""

    saved: Optional[bool] = None

    pipeline_type: Optional[CreateDatasetRequestPipelineType] = None
    r"""The type of data included in the dataset - \"event\" (default) or \"session\" """

    metadata: Optional[Dict[str, Any]] = None
    r"""Any helpful metadata to track for the dataset"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .adddatapoints import (
    AddDatapointsRequest,
    AddDatapointsRequestBody,
    AddDatapointsRequestBodyTypedDict,
    AddDatapointsRequestTypedDict,
    AddDatapointsResponse,
    AddDatapointsResponseBody,
    AddDatapointsResponseBodyTypedDict,
    AddDatapointsResponseTypedDict,
    Mapping,
    MappingTypedDict,
)
from .createconfiguration import (
    CreateConfigurationResponse,
    CreateConfigurationResponseTypedDict,
)
from .createdatapoint import (
    CreateDatapointResponse,
    CreateDatapointResponseBody,
    CreateDatapointResponseBodyTypedDict,
    CreateDatapointResponseTypedDict,
    CreateDatapointResult,
    CreateDatapointResultTypedDict,
)
from .createdataset import (
    CreateDatasetResponse,
    CreateDatasetResponseBody,
    CreateDatasetResponseBodyTypedDict,
    CreateDatasetResponseTypedDict,
    CreateDatasetResult,
    CreateDatasetResultTypedDict,
)
from .createevent import (
    CreateEventRequestBody,
    CreateEventRequestBodyTypedDict,
    CreateEventResponse,
    CreateEventResponseBody,
    CreateEventResponseBodyTypedDict,
    CreateEventResponseTypedDict,
)
from .createeventbatch import (
    CreateEventBatchRequestBody,
    CreateEventBatchRequestBodyTypedDict,
    CreateEventBatchResponse,
    CreateEventBatchResponseBody,
    CreateEventBatchResponseBodyTypedDict,
    CreateEventBatchResponseTypedDict,
)
from .createmetric import CreateMetricResponse, CreateMetricResponseTypedDict
from .createmodelevent import (
    CreateModelEventRequestBody,
    CreateModelEventRequestBodyTypedDict,
    CreateModelEventResponse,
    CreateModelEventResponseBody,
    CreateModelEventResponseBodyTypedDict,
    CreateModelEventResponseTypedDict,
)
from .createmodeleventbatch import (
    CreateModelEventBatchRequestBody,
    CreateModelEventBatchRequestBodyTypedDict,
    CreateModelEventBatchResponse,
    CreateModelEventBatchResponseBody,
    CreateModelEventBatchResponseBodyTypedDict,
    CreateModelEventBatchResponseTypedDict,
)
from .createproject import CreateProjectResponse, CreateProjectResponseTypedDict
from .createrun import CreateRunResponse, CreateRunResponseTypedDict
from .createtool import (
    CreateToolResponse,
    CreateToolResponseBody,
    CreateToolResponseBodyTypedDict,
    CreateToolResponseTypedDict,
    Result,
    ResultTypedDict,
)
from .deleteconfiguration import (
    DeleteConfigurationRequest,
    DeleteConfigurationRequestTypedDict,
    DeleteConfigurationResponse,
    DeleteConfigurationResponseTypedDict,
)
from .deletedatapoint import (
    DeleteDatapointRequest,
    DeleteDatapointRequestTypedDict,
    DeleteDatapointResponse,
    DeleteDatapointResponseBody,
    DeleteDatapointResponseBodyTypedDict,
    DeleteDatapointResponseTypedDict,
)
from .deletedataset import (
    DeleteDatasetRequest,
    DeleteDatasetRequestTypedDict,
    DeleteDatasetResponse,
    DeleteDatasetResponseTypedDict,
)
from .deletemetric import (
    DeleteMetricRequest,
    DeleteMetricRequestTypedDict,
    DeleteMetricResponse,
    DeleteMetricResponseTypedDict,
)
from .deleteproject import (
    DeleteProjectRequest,
    DeleteProjectRequestTypedDict,
    DeleteProjectResponse,
    DeleteProjectResponseTypedDict,
)
from .deleterun import (
    DeleteRunRequest,
    DeleteRunRequestTypedDict,
    DeleteRunResponse,
    DeleteRunResponseTypedDict,
)
from .deletetool import (
    DeleteToolRequest,
    DeleteToolRequestTypedDict,
    DeleteToolResponse,
    DeleteToolResponseTypedDict,
)
from .getconfigurations import (
    Env,
    GetConfigurationsRequest,
    GetConfigurationsRequestTypedDict,
    GetConfigurationsResponse,
    GetConfigurationsResponseTypedDict,
)
from .getdatapoint import (
    GetDatapointRequest,
    GetDatapointRequestTypedDict,
    GetDatapointResponse,
    GetDatapointResponseBody,
    GetDatapointResponseBodyTypedDict,
    GetDatapointResponseTypedDict,
)
from .getdatapoints import (
    GetDatapointsRequest,
    GetDatapointsRequestTypedDict,
    GetDatapointsResponse,
    GetDatapointsResponseBody,
    GetDatapointsResponseBodyTypedDict,
    GetDatapointsResponseTypedDict,
)
from .getdatasets import (
    GetDatasetsRequest,
    GetDatasetsRequestTypedDict,
    GetDatasetsResponse,
    GetDatasetsResponseBody,
    GetDatasetsResponseBodyTypedDict,
    GetDatasetsResponseTypedDict,
    Type,
)
from .getevents import (
    DateRange,
    DateRangeTypedDict,
    GetEventsRequestBody,
    GetEventsRequestBodyTypedDict,
    GetEventsResponse,
    GetEventsResponseBody,
    GetEventsResponseBodyTypedDict,
    GetEventsResponseTypedDict,
)
from .getexperimentcomparison import (
    GetExperimentComparisonRequest,
    GetExperimentComparisonRequestTypedDict,
    GetExperimentComparisonResponse,
    GetExperimentComparisonResponseTypedDict,
    QueryParamAggregateFunction,
)
from .getexperimentresult import (
    AggregateFunction,
    GetExperimentResultRequest,
    GetExperimentResultRequestTypedDict,
    GetExperimentResultResponse,
    GetExperimentResultResponseTypedDict,
)
from .getmetrics import (
    GetMetricsRequest,
    GetMetricsRequestTypedDict,
    GetMetricsResponse,
    GetMetricsResponseTypedDict,
)
from .getprojects import (
    GetProjectsRequest,
    GetProjectsRequestTypedDict,
    GetProjectsResponse,
    GetProjectsResponseTypedDict,
)
from .getrun import (
    GetRunRequest,
    GetRunRequestTypedDict,
    GetRunResponse,
    GetRunResponseTypedDict,
)
from .getruns import (
    GetRunsRequest,
    GetRunsRequestTypedDict,
    GetRunsResponse,
    GetRunsResponseTypedDict,
)
from .getsession import (
    GetSessionRequest,
    GetSessionRequestTypedDict,
    GetSessionResponse,
    GetSessionResponseTypedDict,
)
from .gettools import GetToolsResponse, GetToolsResponseTypedDict
from .startsession import (
    StartSessionRequestBody,
    StartSessionRequestBodyTypedDict,
    StartSessionResponse,
    StartSessionResponseBody,
    StartSessionResponseBodyTypedDict,
    StartSessionResponseTypedDict,
)
from .updateconfiguration import (
    UpdateConfigurationRequest,
    UpdateConfigurationRequestTypedDict,
    UpdateConfigurationResponse,
    UpdateConfigurationResponseTypedDict,
)
from .updatedatapoint import (
    UpdateDatapointRequest,
    UpdateDatapointRequestTypedDict,
    UpdateDatapointResponse,
    UpdateDatapointResponseTypedDict,
)
from .updatedataset import UpdateDatasetResponse, UpdateDatasetResponseTypedDict
from .updateevent import (
    UpdateEventRequestBody,
    UpdateEventRequestBodyTypedDict,
    UpdateEventResponse,
    UpdateEventResponseTypedDict,
)
from .updatemetric import UpdateMetricResponse, UpdateMetricResponseTypedDict
from .updateproject import UpdateProjectResponse, UpdateProjectResponseTypedDict
from .updaterun import (
    UpdateRunRequest,
    UpdateRunRequestTypedDict,
    UpdateRunResponse,
    UpdateRunResponseTypedDict,
)
from .updatetool import UpdateToolResponse, UpdateToolResponseTypedDict

__all__ = [
    "AddDatapointsRequest",
    "AddDatapointsRequestBody",
    "AddDatapointsRequestBodyTypedDict",
    "AddDatapointsRequestTypedDict",
    "AddDatapointsResponse",
    "AddDatapointsResponseBody",
    "AddDatapointsResponseBodyTypedDict",
    "AddDatapointsResponseTypedDict",
    "AggregateFunction",
    "CreateConfigurationResponse",
    "CreateConfigurationResponseTypedDict",
    "CreateDatapointResponse",
    "CreateDatapointResponseBody",
    "CreateDatapointResponseBodyTypedDict",
    "CreateDatapointResponseTypedDict",
    "CreateDatapointResult",
    "CreateDatapointResultTypedDict",
    "CreateDatasetResponse",
    "CreateDatasetResponseBody",
    "CreateDatasetResponseBodyTypedDict",
    "CreateDatasetResponseTypedDict",
    "CreateDatasetResult",
    "CreateDatasetResultTypedDict",
    "CreateEventBatchRequestBody",
    "CreateEventBatchRequestBodyTypedDict",
    "CreateEventBatchResponse",
    "CreateEventBatchResponseBody",
    "CreateEventBatchResponseBodyTypedDict",
    "CreateEventBatchResponseTypedDict",
    "CreateEventRequestBody",
    "CreateEventRequestBodyTypedDict",
    "CreateEventResponse",
    "CreateEventResponseBody",
    "CreateEventResponseBodyTypedDict",
    "CreateEventResponseTypedDict",
    "CreateMetricResponse",
    "CreateMetricResponseTypedDict",
    "CreateModelEventBatchRequestBody",
    "CreateModelEventBatchRequestBodyTypedDict",
    "CreateModelEventBatchResponse",
    "CreateModelEventBatchResponseBody",
    "CreateModelEventBatchResponseBodyTypedDict",
    "CreateModelEventBatchResponseTypedDict",
    "CreateModelEventRequestBody",
    "CreateModelEventRequestBodyTypedDict",
    "CreateModelEventResponse",
    "CreateModelEventResponseBody",
    "CreateModelEventResponseBodyTypedDict",
    "CreateModelEventResponseTypedDict",
    "CreateProjectResponse",
    "CreateProjectResponseTypedDict",
    "CreateRunResponse",
    "CreateRunResponseTypedDict",
    "CreateToolResponse",
    "CreateToolResponseBody",
    "CreateToolResponseBodyTypedDict",
    "CreateToolResponseTypedDict",
    "DateRange",
    "DateRangeTypedDict",
    "DeleteConfigurationRequest",
    "DeleteConfigurationRequestTypedDict",
    "DeleteConfigurationResponse",
    "DeleteConfigurationResponseTypedDict",
    "DeleteDatapointRequest",
    "DeleteDatapointRequestTypedDict",
    "DeleteDatapointResponse",
    "DeleteDatapointResponseBody",
    "DeleteDatapointResponseBodyTypedDict",
    "DeleteDatapointResponseTypedDict",
    "DeleteDatasetRequest",
    "DeleteDatasetRequestTypedDict",
    "DeleteDatasetResponse",
    "DeleteDatasetResponseTypedDict",
    "DeleteMetricRequest",
    "DeleteMetricRequestTypedDict",
    "DeleteMetricResponse",
    "DeleteMetricResponseTypedDict",
    "DeleteProjectRequest",
    "DeleteProjectRequestTypedDict",
    "DeleteProjectResponse",
    "DeleteProjectResponseTypedDict",
    "DeleteRunRequest",
    "DeleteRunRequestTypedDict",
    "DeleteRunResponse",
    "DeleteRunResponseTypedDict",
    "DeleteToolRequest",
    "DeleteToolRequestTypedDict",
    "DeleteToolResponse",
    "DeleteToolResponseTypedDict",
    "Env",
    "GetConfigurationsRequest",
    "GetConfigurationsRequestTypedDict",
    "GetConfigurationsResponse",
    "GetConfigurationsResponseTypedDict",
    "GetDatapointRequest",
    "GetDatapointRequestTypedDict",
    "GetDatapointResponse",
    "GetDatapointResponseBody",
    "GetDatapointResponseBodyTypedDict",
    "GetDatapointResponseTypedDict",
    "GetDatapointsRequest",
    "GetDatapointsRequestTypedDict",
    "GetDatapointsResponse",
    "GetDatapointsResponseBody",
    "GetDatapointsResponseBodyTypedDict",
    "GetDatapointsResponseTypedDict",
    "GetDatasetsRequest",
    "GetDatasetsRequestTypedDict",
    "GetDatasetsResponse",
    "GetDatasetsResponseBody",
    "GetDatasetsResponseBodyTypedDict",
    "GetDatasetsResponseTypedDict",
    "GetEventsRequestBody",
    "GetEventsRequestBodyTypedDict",
    "GetEventsResponse",
    "GetEventsResponseBody",
    "GetEventsResponseBodyTypedDict",
    "GetEventsResponseTypedDict",
    "GetExperimentComparisonRequest",
    "GetExperimentComparisonRequestTypedDict",
    "GetExperimentComparisonResponse",
    "GetExperimentComparisonResponseTypedDict",
    "GetExperimentResultRequest",
    "GetExperimentResultRequestTypedDict",
    "GetExperimentResultResponse",
    "GetExperimentResultResponseTypedDict",
    "GetMetricsRequest",
    "GetMetricsRequestTypedDict",
    "GetMetricsResponse",
    "GetMetricsResponseTypedDict",
    "GetProjectsRequest",
    "GetProjectsRequestTypedDict",
    "GetProjectsResponse",
    "GetProjectsResponseTypedDict",
    "GetRunRequest",
    "GetRunRequestTypedDict",
    "GetRunResponse",
    "GetRunResponseTypedDict",
    "GetRunsRequest",
    "GetRunsRequestTypedDict",
    "GetRunsResponse",
    "GetRunsResponseTypedDict",
    "GetSessionRequest",
    "GetSessionRequestTypedDict",
    "GetSessionResponse",
    "GetSessionResponseTypedDict",
    "GetToolsResponse",
    "GetToolsResponseTypedDict",
    "Mapping",
    "MappingTypedDict",
    "QueryParamAggregateFunction",
    "Result",
    "ResultTypedDict",
    "StartSessionRequestBody",
    "StartSessionRequestBodyTypedDict",
    "StartSessionResponse",
    "StartSessionResponseBody",
    "StartSessionResponseBodyTypedDict",
    "StartSessionResponseTypedDict",
    "Type",
    "UpdateConfigurationRequest",
    "UpdateConfigurationRequestTypedDict",
    "UpdateConfigurationResponse",
    "UpdateConfigurationResponseTypedDict",
    "UpdateDatapointRequest",
    "UpdateDatapointRequestTypedDict",
    "UpdateDatapointResponse",
    "UpdateDatapointResponseTypedDict",
    "UpdateDatasetResponse",
    "UpdateDatasetResponseTypedDict",
    "UpdateEventRequestBody",
    "UpdateEventRequestBodyTypedDict",
    "UpdateEventResponse",
    "UpdateEventResponseTypedDict",
    "UpdateMetricResponse",
    "UpdateMetricResponseTypedDict",
    "UpdateProjectResponse",
    "UpdateProjectResponseTypedDict",
    "UpdateRunRequest",
    "UpdateRunRequestTypedDict",
    "UpdateRunResponse",
    "UpdateRunResponseTypedDict",
    "UpdateToolResponse",
    "UpdateToolResponseTypedDict",
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from enum import Enum
from honeyhive.types import BaseModel
import pydantic
from typing import Any, Dict, List, Optional, TypedDict
from typing_extensions import Annotated, NotRequired


class MetricType(str, Enum):
    r"""Type of the metric - \"custom\", \"model\", \"human\" or \"composite\" """

    CUSTOM = "custom"
    MODEL = "model"
    HUMAN = "human"
    COMPOSITE = "composite"


class ReturnType(str, Enum):
    r"""The data type of the metric value - \"boolean\", \"float\", \"string\" """

    BOOLEAN = "boolean"
    FLOAT = "float"
    STRING = "string"


class ThresholdTypedDict(TypedDict):
    r"""Threshold for numeric metrics to decide passing or failing in tests"""

    min: NotRequired[float]
    max: NotRequired[float]


class Threshold(BaseModel):
    r"""Threshold for numeric metrics to decide passing or failing in tests"""

    min: Optional[float] = None

    max: Optional[float] = None


class MetricTypedDict(TypedDict):
    name: str
    r"""Name of the metric"""
    task: str
    r"""Name of the project associated with metric"""
    type: MetricType
    r"""Type of the metric - \"custom\", \"model\", \"human\" or \"composite\" """
    description: str
    r"""Short description of what the metric does"""
    return_type: ReturnType
    r"""The data type of the metric value - \"boolean\", \"float\", \"string\" """
    criteria: NotRequired[str]
    r"""Criteria for human or composite metrics"""
    code_snippet: NotRequired[str]
    r"""Associated code block for the metric"""
    prompt: NotRequired[str]
    r"""Evaluator prompt for the metric"""
    enabled_in_prod: NotRequired[bool]
    r"""Whether to compute on all production events automatically"""
    needs_ground_truth: NotRequired[bool]
    r"""Whether a ground truth (on metadata) is required to compute it"""
    threshold: NotRequired[ThresholdTypedDict]
    r"""Threshold for numeric metrics to decide passing or failing in tests"""
    pass_when: NotRequired[bool]
    r"""Threshold for boolean metrics to decide passing or failing in tests"""
    id: NotRequired[str]
    r"""Unique idenitifier"""
    event_name: NotRequired[str]
    r"""Name of event that the metric is set to be computed on"""
    event_type: NotRequired[str]
    r"""Type of event that the metric is set to be computed on"""
    model_provider: NotRequired[str]
    r"""Provider of the model, formatted as a LiteLLM provider prefix"""
    model_name: NotRequired[str]
    r"""Name of the model, formatted as a LiteLLM model name"""
    child_metrics: NotRequired[List[Dict[str, Any]]]
    r"""Child metrics added under composite events"""


class Metric(BaseModel):
    name: str
    r"""Name of the metric"""

    task: str
    r"""Name of the project associated with metric"""

    type: MetricType
    r"""Type of the metric - \"custom\", \"model\", \"human\" or \"composite\" """

    description: str
    r"""Short description of what the metric does"""

    return_type: ReturnType
    r"""The data type of the metric value - \"boolean\", \"float\", \"string\" """

    criteria: Optional[str] = None
    r"""Criteria for human or composite metrics"""

    code_snippet: Optional[str] = None
    r"""Associated code block for the metric"""

    prompt: Optional[str] = None
    r"""Evaluator prompt for the metric"""

    enabled_in_prod: Optional[bool] = None
    r"""Whether to compute on all production events automatically"""

    needs_ground_truth: Optional[bool] = None
    r"""Whether a ground truth (on metadata) is required to compute it"""

    threshold: Optional[Threshold] = None
    r"""Threshold for numeric metrics to decide passing or failing in tests"""

    pass_when: Optional[bool] = None
    r"""Threshold for boolean metrics to decide passing or failing in tests"""

    id: Annotated[Optional[str], pydantic.Field(alias="_id")] = None
    r"""Unique idenitifier"""

    event_name: Optional[str] = None
    r"""Name of event that the metric is set to be computed on"""

    event_type: Optional[str] = None
    r"""Type of event that the metric is set to be computed on"""

    model_provider: Optional[str] = None
    r"""Provider of the model, formatted as a LiteLLM provider prefix"""

    model_name: Optional[str] = None
    r"""Name of the model, formatted as a LiteLLM model name"""

    child_metrics: Optional[List[Dict[str, Any]]] = None
    r"""Child metrics added under composite events"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from honeyhive.types import BaseModel
from typing import Optional, TypedDict
from typing_extensions import NotRequired


class ProjectTypedDict(TypedDict):
    name: str
    description: str
    id: NotRequired[str]


class Project(BaseModel):
    name: str

    description: str

    id: Optional[str] = None



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from enum import Enum
from honeyhive.models.components import (
    experimentcomparisonresponse as components_experimentcomparisonresponse,
)
from honeyhive.types import BaseModel
from honeyhive.utils import FieldMetadata, PathParamMetadata, QueryParamMetadata
import httpx
from typing import Optional, TypedDict
from typing_extensions import Annotated, NotRequired


class QueryParamAggregateFunction(str, Enum):
    AVERAGE = "average"
    MIN = "min"
    MAX = "max"
    MEDIAN = "median"
    P95 = "p95"
    P99 = "p99"
    P90 = "p90"
    SUM = "sum"
    COUNT = "count"


class GetExperimentComparisonRequestTypedDict(TypedDict):
    run_id_1: str
    run_id_2: str
    project_id: str
    aggregate_function: NotRequired[QueryParamAggregateFunction]


class GetExperimentComparisonRequest(BaseModel):
    run_id_1: Annotated[
        str, FieldMetadata(path=PathParamMetadata(style="simple", explode=False))
    ]

    run_id_2: Annotated[
        str, FieldMetadata(path=PathParamMetadata(style="simple", explode=False))
    ]

    project_id: Annotated[
        str, FieldMetadata(query=QueryParamMetadata(style="form", explode=True))
    ]

    aggregate_function: Annotated[
        Optional[QueryParamAggregateFunction],
        FieldMetadata(query=QueryParamMetadata(style="form", explode=True)),
    ] = None


class GetExperimentComparisonResponseTypedDict(TypedDict):
    content_type: str
    r"""HTTP response content type for this operation"""
    status_code: int
    r"""HTTP response status code for this operation"""
    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""
    experiment_comparison_response: NotRequired[
        components_experimentcomparisonresponse.ExperimentComparisonResponseTypedDict
    ]
    r"""Experiment comparison retrieved successfully"""


class GetExperimentComparisonResponse(BaseModel):
    content_type: str
    r"""HTTP response content type for this operation"""

    status_code: int
    r"""HTTP response status code for this operation"""

    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""

    experiment_comparison_response: Optional[
        components_experimentcomparisonresponse.ExperimentComparisonResponse
    ] = None
    r"""Experiment comparison retrieved successfully"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from honeyhive.models.components import (
    createeventrequest as components_createeventrequest,
    sessionpropertiesbatch as components_sessionpropertiesbatch,
)
from honeyhive.types import BaseModel
import httpx
from typing import List, Optional, TypedDict
from typing_extensions import NotRequired


class CreateEventBatchRequestBodyTypedDict(TypedDict):
    events: List[components_createeventrequest.CreateEventRequestTypedDict]
    is_single_session: NotRequired[bool]
    r"""Default is false. If true, all events will be associated with the same session"""
    session_properties: NotRequired[
        components_sessionpropertiesbatch.SessionPropertiesBatchTypedDict
    ]


class CreateEventBatchRequestBody(BaseModel):
    events: List[components_createeventrequest.CreateEventRequest]

    is_single_session: Optional[bool] = None
    r"""Default is false. If true, all events will be associated with the same session"""

    session_properties: Optional[
        components_sessionpropertiesbatch.SessionPropertiesBatch
    ] = None


class CreateEventBatchResponseBodyTypedDict(TypedDict):
    r"""Events created"""

    event_ids: NotRequired[List[str]]
    session_id: NotRequired[str]
    success: NotRequired[bool]


class CreateEventBatchResponseBody(BaseModel):
    r"""Events created"""

    event_ids: Optional[List[str]] = None

    session_id: Optional[str] = None

    success: Optional[bool] = None


class CreateEventBatchResponseTypedDict(TypedDict):
    content_type: str
    r"""HTTP response content type for this operation"""
    status_code: int
    r"""HTTP response status code for this operation"""
    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""
    object: NotRequired[CreateEventBatchResponseBodyTypedDict]
    r"""Events created"""


class CreateEventBatchResponse(BaseModel):
    content_type: str
    r"""HTTP response content type for this operation"""

    status_code: int
    r"""HTTP response status code for this operation"""

    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""

    object: Optional[CreateEventBatchResponseBody] = None
    r"""Events created"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from honeyhive.types import BaseModel
from typing import Any, Dict, Optional, TypedDict
from typing_extensions import NotRequired


class SessionPropertiesBatchTypedDict(TypedDict):
    session_name: NotRequired[str]
    r"""Name of the session"""
    source: NotRequired[str]
    r"""Source of the session - production, staging, etc"""
    session_id: NotRequired[str]
    r"""Unique id of the session, if not set, it will be auto-generated"""
    config: NotRequired[Dict[str, Any]]
    r"""Associated configuration for the session"""
    inputs: NotRequired[Dict[str, Any]]
    r"""Input object passed to the session - user query, text blob, etc"""
    outputs: NotRequired[Dict[str, Any]]
    r"""Final output of the session - completion, chunks, etc"""
    error: NotRequired[str]
    r"""Any error description if session failed"""
    user_properties: NotRequired[Dict[str, Any]]
    r"""Any user properties associated with the session"""
    metrics: NotRequired[Dict[str, Any]]
    r"""Any values computed over the output of the session"""
    feedback: NotRequired[Dict[str, Any]]
    r"""Any user feedback provided for the session output"""
    metadata: NotRequired[Dict[str, Any]]
    r"""Any system or application metadata associated with the session"""


class SessionPropertiesBatch(BaseModel):
    session_name: Optional[str] = None
    r"""Name of the session"""

    source: Optional[str] = None
    r"""Source of the session - production, staging, etc"""

    session_id: Optional[str] = None
    r"""Unique id of the session, if not set, it will be auto-generated"""

    config: Optional[Dict[str, Any]] = None
    r"""Associated configuration for the session"""

    inputs: Optional[Dict[str, Any]] = None
    r"""Input object passed to the session - user query, text blob, etc"""

    outputs: Optional[Dict[str, Any]] = None
    r"""Final output of the session - completion, chunks, etc"""

    error: Optional[str] = None
    r"""Any error description if session failed"""

    user_properties: Optional[Dict[str, Any]] = None
    r"""Any user properties associated with the session"""

    metrics: Optional[Dict[str, Any]] = None
    r"""Any values computed over the output of the session"""

    feedback: Optional[Dict[str, Any]] = None
    r"""Any user feedback provided for the session output"""

    metadata: Optional[Dict[str, Any]] = None
    r"""Any system or application metadata associated with the session"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .configuration import (
    CallType,
    Configuration,
    ConfigurationType,
    ConfigurationTypedDict,
    Env,
    FunctionCallParams,
    Parameters,
    ParametersTypedDict,
    ResponseFormat,
    ResponseFormatTypedDict,
    SelectedFunctions,
    SelectedFunctionsTypedDict,
)
from .createdatapointrequest import (
    CreateDatapointRequest,
    CreateDatapointRequestTypedDict,
)
from .createdatasetrequest import (
    CreateDatasetRequest,
    CreateDatasetRequestPipelineType,
    CreateDatasetRequestType,
    CreateDatasetRequestTypedDict,
)
from .createeventrequest import (
    CreateEventRequest,
    CreateEventRequestEventType,
    CreateEventRequestTypedDict,
)
from .createmodelevent import CreateModelEvent, CreateModelEventTypedDict
from .createprojectrequest import CreateProjectRequest, CreateProjectRequestTypedDict
from .createrunrequest import CreateRunRequest, CreateRunRequestTypedDict, Status
from .createrunresponse import CreateRunResponse, CreateRunResponseTypedDict
from .createtoolrequest import (
    CreateToolRequest,
    CreateToolRequestType,
    CreateToolRequestTypedDict,
)
from .datapoint import Datapoint, DatapointTypedDict
from .dataset import Dataset, DatasetType, DatasetTypedDict, PipelineType
from .datasetupdate import DatasetUpdate, DatasetUpdateTypedDict
from .deleterunresponse import DeleteRunResponse, DeleteRunResponseTypedDict
from .evaluationrun import (
    EvaluationRun,
    EvaluationRunStatus,
    EvaluationRunTypedDict,
    Results,
    ResultsTypedDict,
)
from .event import Event, EventType, EventTypedDict
from .eventfilter import EventFilter, EventFilterTypedDict, Operator, Type
from .experimentcomparisonresponse import (
    Evaluators,
    EvaluatorsTypedDict,
    EventDetails,
    EventDetailsTypedDict,
    ExperimentComparisonResponse,
    ExperimentComparisonResponseConfiguration,
    ExperimentComparisonResponseConfigurationTypedDict,
    ExperimentComparisonResponseEvaluators,
    ExperimentComparisonResponseEvaluatorsTypedDict,
    ExperimentComparisonResponseMetadata,
    ExperimentComparisonResponseMetadataTypedDict,
    ExperimentComparisonResponseMetrics,
    ExperimentComparisonResponseMetricsTypedDict,
    ExperimentComparisonResponsePassingRanges,
    ExperimentComparisonResponsePassingRangesTypedDict,
    ExperimentComparisonResponseResults,
    ExperimentComparisonResponseResultsTypedDict,
    ExperimentComparisonResponseSchemasConfiguration,
    ExperimentComparisonResponseSchemasConfigurationTypedDict,
    ExperimentComparisonResponseSchemasResults,
    ExperimentComparisonResponseSchemasResultsTypedDict,
    ExperimentComparisonResponseTypedDict,
    Metadata,
    MetadataTypedDict,
    NewRun,
    NewRunTypedDict,
    NewValues,
    NewValuesTypedDict,
    OldRun,
    OldRunTypedDict,
    OldValues,
    OldValuesTypedDict,
    PassingRanges,
    PassingRangesTypedDict,
)
from .experimentresultresponse import (
    Datapoints,
    DatapointsTypedDict,
    Details,
    DetailsTypedDict,
    ExperimentResultResponse,
    ExperimentResultResponseDatapoints,
    ExperimentResultResponseDatapointsTypedDict,
    ExperimentResultResponseMetrics,
    ExperimentResultResponseMetricsTypedDict,
    ExperimentResultResponseTypedDict,
    Metrics,
    MetricsTypedDict,
    Value,
    ValueTypedDict,
    Values,
    ValuesTypedDict,
)
from .getrunresponse import GetRunResponse, GetRunResponseTypedDict
from .getrunsresponse import GetRunsResponse, GetRunsResponseTypedDict
from .metric import (
    Metric,
    MetricType,
    MetricTypedDict,
    ReturnType,
    Threshold,
    ThresholdTypedDict,
)
from .metricedit import (
    MetricEdit,
    MetricEditEventType,
    MetricEditReturnType,
    MetricEditThreshold,
    MetricEditThresholdTypedDict,
    MetricEditType,
    MetricEditTypedDict,
)
from .postconfigurationrequest import (
    PostConfigurationRequest,
    PostConfigurationRequestCallType,
    PostConfigurationRequestEnv,
    PostConfigurationRequestFunctionCallParams,
    PostConfigurationRequestParameters,
    PostConfigurationRequestParametersTypedDict,
    PostConfigurationRequestResponseFormat,
    PostConfigurationRequestResponseFormatTypedDict,
    PostConfigurationRequestSelectedFunctions,
    PostConfigurationRequestSelectedFunctionsTypedDict,
    PostConfigurationRequestTypedDict,
)
from .project import Project, ProjectTypedDict
from .putconfigurationrequest import (
    PutConfigurationRequest,
    PutConfigurationRequestCallType,
    PutConfigurationRequestEnv,
    PutConfigurationRequestFunctionCallParams,
    PutConfigurationRequestParameters,
    PutConfigurationRequestParametersTypedDict,
    PutConfigurationRequestResponseFormat,
    PutConfigurationRequestResponseFormatTypedDict,
    PutConfigurationRequestSelectedFunctions,
    PutConfigurationRequestSelectedFunctionsTypedDict,
    PutConfigurationRequestType,
    PutConfigurationRequestTypedDict,
)
from .security import Security, SecurityTypedDict
from .sessionpropertiesbatch import (
    SessionPropertiesBatch,
    SessionPropertiesBatchTypedDict,
)
from .sessionstartrequest import SessionStartRequest, SessionStartRequestTypedDict
from .tool import Tool, ToolType, ToolTypedDict
from .updatedatapointrequest import (
    UpdateDatapointRequest,
    UpdateDatapointRequestTypedDict,
)
from .updateprojectrequest import UpdateProjectRequest, UpdateProjectRequestTypedDict
from .updaterunrequest import (
    UpdateRunRequest,
    UpdateRunRequestStatus,
    UpdateRunRequestTypedDict,
)
from .updaterunresponse import UpdateRunResponse, UpdateRunResponseTypedDict
from .updatetoolrequest import UpdateToolRequest, UpdateToolRequestTypedDict

__all__ = [
    "CallType",
    "Configuration",
    "ConfigurationType",
    "ConfigurationTypedDict",
    "CreateDatapointRequest",
    "CreateDatapointRequestTypedDict",
    "CreateDatasetRequest",
    "CreateDatasetRequestPipelineType",
    "CreateDatasetRequestType",
    "CreateDatasetRequestTypedDict",
    "CreateEventRequest",
    "CreateEventRequestEventType",
    "CreateEventRequestTypedDict",
    "CreateModelEvent",
    "CreateModelEventTypedDict",
    "CreateProjectRequest",
    "CreateProjectRequestTypedDict",
    "CreateRunRequest",
    "CreateRunRequestTypedDict",
    "CreateRunResponse",
    "CreateRunResponseTypedDict",
    "CreateToolRequest",
    "CreateToolRequestType",
    "CreateToolRequestTypedDict",
    "Datapoint",
    "DatapointTypedDict",
    "Datapoints",
    "DatapointsTypedDict",
    "Dataset",
    "DatasetType",
    "DatasetTypedDict",
    "DatasetUpdate",
    "DatasetUpdateTypedDict",
    "DeleteRunResponse",
    "DeleteRunResponseTypedDict",
    "Details",
    "DetailsTypedDict",
    "Env",
    "EvaluationRun",
    "EvaluationRunStatus",
    "EvaluationRunTypedDict",
    "Evaluators",
    "EvaluatorsTypedDict",
    "Event",
    "EventDetails",
    "EventDetailsTypedDict",
    "EventFilter",
    "EventFilterTypedDict",
    "EventType",
    "EventTypedDict",
    "ExperimentComparisonResponse",
    "ExperimentComparisonResponseConfiguration",
    "ExperimentComparisonResponseConfigurationTypedDict",
    "ExperimentComparisonResponseEvaluators",
    "ExperimentComparisonResponseEvaluatorsTypedDict",
    "ExperimentComparisonResponseMetadata",
    "ExperimentComparisonResponseMetadataTypedDict",
    "ExperimentComparisonResponseMetrics",
    "ExperimentComparisonResponseMetricsTypedDict",
    "ExperimentComparisonResponsePassingRanges",
    "ExperimentComparisonResponsePassingRangesTypedDict",
    "ExperimentComparisonResponseResults",
    "ExperimentComparisonResponseResultsTypedDict",
    "ExperimentComparisonResponseSchemasConfiguration",
    "ExperimentComparisonResponseSchemasConfigurationTypedDict",
    "ExperimentComparisonResponseSchemasResults",
    "ExperimentComparisonResponseSchemasResultsTypedDict",
    "ExperimentComparisonResponseTypedDict",
    "ExperimentResultResponse",
    "ExperimentResultResponseDatapoints",
    "ExperimentResultResponseDatapointsTypedDict",
    "ExperimentResultResponseMetrics",
    "ExperimentResultResponseMetricsTypedDict",
    "ExperimentResultResponseTypedDict",
    "FunctionCallParams",
    "GetRunResponse",
    "GetRunResponseTypedDict",
    "GetRunsResponse",
    "GetRunsResponseTypedDict",
    "Metadata",
    "MetadataTypedDict",
    "Metric",
    "MetricEdit",
    "MetricEditEventType",
    "MetricEditReturnType",
    "MetricEditThreshold",
    "MetricEditThresholdTypedDict",
    "MetricEditType",
    "MetricEditTypedDict",
    "MetricType",
    "MetricTypedDict",
    "Metrics",
    "MetricsTypedDict",
    "NewRun",
    "NewRunTypedDict",
    "NewValues",
    "NewValuesTypedDict",
    "OldRun",
    "OldRunTypedDict",
    "OldValues",
    "OldValuesTypedDict",
    "Operator",
    "Parameters",
    "ParametersTypedDict",
    "PassingRanges",
    "PassingRangesTypedDict",
    "PipelineType",
    "PostConfigurationRequest",
    "PostConfigurationRequestCallType",
    "PostConfigurationRequestEnv",
    "PostConfigurationRequestFunctionCallParams",
    "PostConfigurationRequestParameters",
    "PostConfigurationRequestParametersTypedDict",
    "PostConfigurationRequestResponseFormat",
    "PostConfigurationRequestResponseFormatTypedDict",
    "PostConfigurationRequestSelectedFunctions",
    "PostConfigurationRequestSelectedFunctionsTypedDict",
    "PostConfigurationRequestTypedDict",
    "Project",
    "ProjectTypedDict",
    "PutConfigurationRequest",
    "PutConfigurationRequestCallType",
    "PutConfigurationRequestEnv",
    "PutConfigurationRequestFunctionCallParams",
    "PutConfigurationRequestParameters",
    "PutConfigurationRequestParametersTypedDict",
    "PutConfigurationRequestResponseFormat",
    "PutConfigurationRequestResponseFormatTypedDict",
    "PutConfigurationRequestSelectedFunctions",
    "PutConfigurationRequestSelectedFunctionsTypedDict",
    "PutConfigurationRequestType",
    "PutConfigurationRequestTypedDict",
    "ResponseFormat",
    "ResponseFormatTypedDict",
    "Results",
    "ResultsTypedDict",
    "ReturnType",
    "Security",
    "SecurityTypedDict",
    "SelectedFunctions",
    "SelectedFunctionsTypedDict",
    "SessionPropertiesBatch",
    "SessionPropertiesBatchTypedDict",
    "SessionStartRequest",
    "SessionStartRequestTypedDict",
    "Status",
    "Threshold",
    "ThresholdTypedDict",
    "Tool",
    "ToolType",
    "ToolTypedDict",
    "Type",
    "UpdateDatapointRequest",
    "UpdateDatapointRequestTypedDict",
    "UpdateProjectRequest",
    "UpdateProjectRequestTypedDict",
    "UpdateRunRequest",
    "UpdateRunRequestStatus",
    "UpdateRunRequestTypedDict",
    "UpdateRunResponse",
    "UpdateRunResponseTypedDict",
    "UpdateToolRequest",
    "UpdateToolRequestTypedDict",
    "Value",
    "ValueTypedDict",
    "Values",
    "ValuesTypedDict",
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from enum import Enum
from honeyhive.types import BaseModel
from typing import List, Optional, TypedDict
from typing_extensions import NotRequired


class DatasetType(str, Enum):
    r"""What the dataset is to be used for - \"evaluation\" or \"fine-tuning\" """

    EVALUATION = "evaluation"
    FINE_TUNING = "fine-tuning"


class PipelineType(str, Enum):
    r"""The type of data included in the dataset - \"event\" (default) or \"session\" """

    EVENT = "event"
    SESSION = "session"


class DatasetTypedDict(TypedDict):
    project: NotRequired[str]
    r"""UUID of the project associated with this dataset"""
    name: NotRequired[str]
    r"""Name of the dataset"""
    description: NotRequired[str]
    r"""A description for the dataset"""
    type: NotRequired[DatasetType]
    r"""What the dataset is to be used for - \"evaluation\" or \"fine-tuning\" """
    datapoints: NotRequired[List[str]]
    r"""List of unique datapoint ids to be included in this dataset"""
    num_points: NotRequired[int]
    r"""Number of datapoints included in the dataset"""
    linked_evals: NotRequired[List[str]]
    saved: NotRequired[bool]
    r"""Whether the dataset has been saved or detected"""
    pipeline_type: NotRequired[PipelineType]
    r"""The type of data included in the dataset - \"event\" (default) or \"session\" """
    created_at: NotRequired[str]
    r"""Timestamp of when the dataset was created"""
    updated_at: NotRequired[str]
    r"""Timestamp of when the dataset was last updated"""


class Dataset(BaseModel):
    project: Optional[str] = None
    r"""UUID of the project associated with this dataset"""

    name: Optional[str] = None
    r"""Name of the dataset"""

    description: Optional[str] = None
    r"""A description for the dataset"""

    type: Optional[DatasetType] = None
    r"""What the dataset is to be used for - \"evaluation\" or \"fine-tuning\" """

    datapoints: Optional[List[str]] = None
    r"""List of unique datapoint ids to be included in this dataset"""

    num_points: Optional[int] = None
    r"""Number of datapoints included in the dataset"""

    linked_evals: Optional[List[str]] = None

    saved: Optional[bool] = None
    r"""Whether the dataset has been saved or detected"""

    pipeline_type: Optional[PipelineType] = None
    r"""The type of data included in the dataset - \"event\" (default) or \"session\" """

    created_at: Optional[str] = None
    r"""Timestamp of when the dataset was created"""

    updated_at: Optional[str] = None
    r"""Timestamp of when the dataset was last updated"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .evaluationrun import EvaluationRun, EvaluationRunTypedDict
from honeyhive.types import BaseModel
from typing import Optional, TypedDict
from typing_extensions import NotRequired


class CreateRunResponseTypedDict(TypedDict):
    evaluation: NotRequired[EvaluationRunTypedDict]
    run_id: NotRequired[str]


class CreateRunResponse(BaseModel):
    evaluation: Optional[EvaluationRun] = None

    run_id: Optional[str] = None



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .sdk import *
from .sdkconfiguration import *
from .tracer import HoneyHiveTracer, enrich_session 
from .tracer.custom import trace, atrace, enrich_span
from .evaluation import evaluate, evaluator, aevaluator
from .utils.dotdict import dotdict
from .utils.config import config

__all__ = [
    "HoneyHiveTracer",
    "enrich_session",
    "trace",
    "atrace",
    "enrich_span",
    "evaluate",
    "evaluator",
    "aevaluator",
    "dotdict",
    "config"
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from honeyhive.types import BaseModel
from typing import Optional, TypedDict
from typing_extensions import NotRequired


class UpdateProjectRequestTypedDict(TypedDict):
    project_id: str
    name: NotRequired[str]
    description: NotRequired[str]


class UpdateProjectRequest(BaseModel):
    project_id: str

    name: Optional[str] = None

    description: Optional[str] = None



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from honeyhive.models.components import getrunsresponse as components_getrunsresponse
from honeyhive.types import BaseModel
from honeyhive.utils import FieldMetadata, QueryParamMetadata
import httpx
from typing import Optional, TypedDict
from typing_extensions import Annotated, NotRequired


class GetRunsRequestTypedDict(TypedDict):
    project: NotRequired[str]


class GetRunsRequest(BaseModel):
    project: Annotated[
        Optional[str],
        FieldMetadata(query=QueryParamMetadata(style="form", explode=True)),
    ] = None


class GetRunsResponseTypedDict(TypedDict):
    content_type: str
    r"""HTTP response content type for this operation"""
    status_code: int
    r"""HTTP response status code for this operation"""
    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""
    get_runs_response: NotRequired[components_getrunsresponse.GetRunsResponseTypedDict]
    r"""Successful response"""


class GetRunsResponse(BaseModel):
    content_type: str
    r"""HTTP response content type for this operation"""

    status_code: int
    r"""HTTP response status code for this operation"""

    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""

    get_runs_response: Optional[components_getrunsresponse.GetRunsResponse] = None
    r"""Successful response"""



from honeyhive.sdk import HoneyHive
from honeyhive.models import components
from honeyhive import HoneyHiveTracer, enrich_session
from .evaluators import evaluator, aevaluator

from concurrent.futures import ThreadPoolExecutor
import collections
import contextvars
import functools
import datetime

from rich.style import Style
from rich.console import Console
from rich.table import Table

from dataclasses import dataclass
from typing import Optional, List, Dict, Any, Callable, Union
import os
import hashlib
import json
import time
import sys
import traceback
import asyncio
import inspect

from opentelemetry import context
from opentelemetry.context import Context

@dataclass
class EvaluationResult:
    run_id: str
    stats: Dict[str, Any]
    dataset_id: str 
    session_ids: list
    status: str
    suite: str
    data: Dict[str, list]

    def to_json(self):
        # save data dict to json file
        with open(f"{self.suite}.json", "w") as f:
            json.dump(self.data, f, indent=4)

console = Console()

class Evaluation:
    """This class is for automated honeyhive evaluation with tracing"""

    def __init__(
        self,
        api_key: str = None,
        project: str = None,
        name: Optional[str] = None,
        suite: Optional[str] = None,
        function: Optional[Callable] = None,
        dataset: Optional[Union[List[Any], Dict[str, Any]]] = None,
        evaluators: Optional[List[Any]] = None,
        dataset_id: Optional[str] = None,
        max_workers: int = 10,
        run_concurrently: bool = True,
        disable_http_tracing: bool = False,
        server_url: Optional[str] = None,
        verbose: bool = False,
        metadata: Optional[Dict[str, Any]] = None,
        print_results: bool = True,
    ):
        if HoneyHiveTracer._is_traceloop_initialized:
            # undo traceloop initialization
            ctx: Context = context.get_current()
            ctx = context.set_value('association_properties', None)
            context.attach(ctx)
            HoneyHiveTracer._is_traceloop_initialized = False
        
        if function is None:
            raise Exception(
                "Please provide a function to evaluate."
            )
        
        # if name is not provided, use the file name of the caller
        try:
            if name is None:
                # Get the frame of the caller (1 level up in the stack)
                caller_frame = sys._getframe(2)
                # Extract the filename of the caller
                name = os.path.basename(caller_frame.f_code.co_filename)
        except Exception:
            current_mm_dd = datetime.datetime.now().strftime("%m/%d")
            name = f"Offline Eval {current_mm_dd}"

        # get the directory of the file being evaluated
        try:
            if suite is None:
                suite = os.path.dirname(sys._getframe(1).f_code.co_filename).split(os.sep)[-1]
        except Exception:
            suite = "default"
        
        self.api_key = api_key or os.environ["HH_API_KEY"]
        self.project = project or os.environ["HH_PROJECT"]
        self.name: str = name 
        self.evaluators = evaluators or []
        self.status: str = "pending"
        self.max_workers: int = max(1, max_workers if max_workers else int(os.getenv("HH_MAX_WORKERS", 10)))
        self.run_concurrently: bool = run_concurrently
        self.function: Callable = function
        self.suite = suite
        self.disable_auto_tracing = True
        self.eval_run: Optional[components.CreateRunResponse] = None
        self.evaluation_session_ids: collections.deque = collections.deque()
        self.server_url = server_url or os.environ.get("HH_API_URL", "https://api.honeyhive.ai")
        self.verbose = verbose
        self.disable_http_tracing = disable_http_tracing
        self.metadata = metadata or {}
        self.print_results = print_results

        # only one of dataset_id or dataset should be provided
        self.dataset_id: Optional[str] = dataset_id
        self.dataset: Optional[List[Any]] = dataset if isinstance(dataset, list) else dataset["data"] if isinstance(dataset, dict) else None
        self.external_dataset_params: Optional[Dict[str, str]] = {
            "id": dataset["id"],
            "name": dataset.get("name"),
        } if isinstance(dataset, dict) and "id" in dataset else None
        
        if not self.dataset_id and not self.dataset:
            raise Exception(
                "No valid 'dataset_id' or 'dataset' found. Please provide one to iterate the evaluation over."
            )
        elif self.dataset_id and self.dataset:
            raise Exception(
                "Both 'dataset_id' and 'dataset' were provided. Please provide only one of them for evaluation."
            )
        self.use_hh_dataset: bool = self.dataset_id is not None

        # Add git information to metadata if available
        try:
            git_info = HoneyHiveTracer._get_git_info()
            if "error" not in git_info:
                self.metadata["git"] = git_info
        except Exception as e:
            if self.verbose:
                print(f"Error getting git info: {e}")

        if self.external_dataset_params and self.external_dataset_params.get('name'): 
            self.metadata['dataset_name'] = self.external_dataset_params['name']

        self._validate_requirements()
        self.hhai = HoneyHive(bearer_auth=self.api_key, server_url=server_url)

        self._setup_dataset()

        # increase the OTEL export timeout to 30 seconds
        # os.environ["OTEL_EXPORTER_OTLP_TIMEOUT"] = "30000"

        # mark the tracer for evaluation tracing
        HoneyHiveTracer.is_evaluation = True

    def _setup_dataset(self) -> None:
        """
        Set up the dataset for evaluation:
        - Loads dataset from HoneyHive using dataset_id if provided
        - Uses provided dataset list if passed directly
        - Generates an external dataset ID for custom datasets
        - Validates dataset format (must be a list of dictionaries)
        - Raises exceptions if no dataset is provided or format is invalid
        """

        assert (self.dataset_id is None) != (self.dataset is None), "Either dataset_id or dataset must be defined but not both"
        
        # load the dataset from HoneyHive
        if self.use_hh_dataset:
            try:
                dataset = self.hhai.datasets.get_datasets(
                    project=self.project,
                    dataset_id=self.dataset_id,
                )
                if (
                    dataset
                    and dataset.object.testcases
                    and len(dataset.object.testcases) > 0
                ):
                    self.dataset = dataset.object.testcases[0]
                else:
                    raise RuntimeError(f"No valid testcases found in dataset {self.dataset_id}")
            except Exception:
                raise RuntimeError(
                    f"No dataset found with id - {self.dataset_id} for project - {self.project}. Please use the correct dataset id and project name."
                )
        # use provided dataset
        else:

            # validate dataset format
            if not isinstance(self.dataset, list):
                raise Exception("Dataset must be a list")
            if not all(isinstance(item, dict) for item in self.dataset):
                raise Exception("All items in dataset must be dictionaries")            

            # generated id for external datasets
            # TODO: large dataset optimization
            # TODO: dataset might not be json serializable
            self.dataset_id: str = (
                self._add_ext_prefix(self.external_dataset_params["id"]) if self.external_dataset_params and "id" in self.external_dataset_params
                else Evaluation.generate_hash(json.dumps(self.dataset)) if self.dataset 
                else None
            )
        assert self.dataset_id is not None and self.dataset is not None

    def _validate_requirements(self) -> None:
        """Sanity check of requirements for HoneyHive evaluations and tracing."""
        if not self.api_key:
            raise Exception(
                "Honeyhive API key not found. Please set 'api_key' to initiate Honeyhive Tracer. Cannot run Evaluation"
            )
        if not self.project:
            raise Exception(
                "Honeyhive Project not found. Please set 'project' to initiate Honeyhive Tracer. Cannot run Evaluation"
            )

    @staticmethod
    def _add_ext_prefix(id_string) -> str:
        """Add EXT- prefix to an ID if it doesn't already have it"""
        if not isinstance(id_string, str):
            id_string = str(id_string)
        if not id_string.startswith("EXT-"):
            return f"EXT-{id_string}"
        return id_string

    @staticmethod
    def generate_hash(input_string: str) -> str:
        return Evaluation._add_ext_prefix(hashlib.md5(input_string.encode('utf-8')).hexdigest()[:24])

    # ------------------------------------------------------------

    def _get_tracing_metadata(
        self,
        datapoint_idx: int,
    ):
        """Get tracing metadata for evaluation."""
        tracing_metadata = {"run_id": self.eval_run.run_id}
        if self.use_hh_dataset:
            datapoint_id = self.dataset.datapoints[datapoint_idx]
            # Convert to string if it's an integer
            if isinstance(datapoint_id, int):
                datapoint_id = str(datapoint_id)
            tracing_metadata["datapoint_id"] = datapoint_id
        else:
            tracing_metadata["datapoint_id"] = (
                self._add_ext_prefix(self.dataset[datapoint_idx]["id"]) if isinstance(self.dataset[datapoint_idx], dict) and "id" in self.dataset[datapoint_idx]
                else Evaluation.generate_hash(json.dumps(self.dataset[datapoint_idx]))
            )
        
        tracing_metadata["dataset_id"] = self.dataset_id

        return tracing_metadata

    def _enrich_evaluation_session(
        self,
        datapoint_idx: int,
        session_id: str,
        outputs: Optional[Any],
        metrics: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        """Private function to enrich the session data post flow completion."""
        try:
            tracing_metadata = self._get_tracing_metadata(datapoint_idx)
            tracing_metadata.update(metadata)

            if not isinstance(outputs, dict):
                outputs = {"output": outputs}

            enrich_session(
                session_id=session_id,
                metadata=tracing_metadata,
                outputs=outputs,
                metrics=metrics,
            )
        except Exception as e:
            print(f"Error adding trace metadata: {e}")

    def _get_evaluator_metadata(self, eval_func, evaluator_name: str) -> dict:
        """Get metadata for an evaluator if it's decorated."""
        if not isinstance(eval_func, evaluator):
            return {}
            
        eval_settings_dict = evaluator.all_evaluator_settings[evaluator_name].resolve_settings().dict()
        # remove all None values, weight if 1.0 and asserts if False
        filtered_dict = {}
        for k, v in eval_settings_dict.items():
            if not (v is None or (k == 'weight' and v == 1.0) or (k == 'asserts' and v is False)):
                filtered_dict[k] = v
        return {
            'eval_settings': filtered_dict,
        }

    def _run_single_evaluator(
        self,
        eval_func,
        evaluator_name: str,
        outputs: Any = None,
        inputs: Dict[str, Any] = None,
        ground_truth: Dict[str, Any] = None
    ) -> tuple[str, Any, dict]:
        """Run a single evaluator and return its name, result and metadata."""
        metadata = self._get_evaluator_metadata(eval_func, evaluator_name)

        try:
            evaluator_result = eval_func(outputs, inputs, ground_truth)
            return evaluator_name, evaluator_result, metadata

        except Exception as e:
            print(f"Error in evaluator: {str(e)}\n")
            print(traceback.format_exc())
            return evaluator_name, None, metadata

    async def _arun_single_evaluator(
       self,
        eval_func,
        evaluator_name: str,
        outputs: Any = None,
        inputs: Dict[str, Any] = None,
        ground_truth: Dict[str, Any] = None
    ) -> tuple[str, Any, dict]:
        """Run a single async evaluator and return its name, result and metadata."""
        metadata = self._get_evaluator_metadata(eval_func, evaluator_name)

        try:
            evaluator_result = await eval_func(outputs, inputs, ground_truth)
            return evaluator_name, evaluator_result, metadata

        except Exception as e:
            print(f"Error in evaluator: {str(e)}\n")
            print(traceback.format_exc())
            return evaluator_name, None, metadata

    def _run_evaluators(
        self, 
        outputs: Optional[Any], 
        inputs: Optional[Dict[str, Any]], 
        ground_truth: Optional[Dict[str, Any]]
    ):
        """Run evaluators and collect metrics."""
        metrics = {}
        metadata = {}

        if not self.evaluators:
            return metrics, metadata

        # Separate sync and async evaluators
        sync_evaluators = []
        async_evaluators = []
        eval_names = set()

        for index, eval_func in enumerate(self.evaluators):
            evaluator_name = getattr(eval_func, "__name__", f"evaluator_{index}")
            if evaluator_name in eval_names:
                raise ValueError(f"Evaluator {evaluator_name} is defined multiple times")
            eval_names.add(evaluator_name)

            if inspect.iscoroutinefunction(eval_func):
                async_evaluators.append((eval_func, evaluator_name))
            else:
                sync_evaluators.append((eval_func, evaluator_name))

        # Run sync evaluators first
        for eval_func, name in sync_evaluators:
            name, result, meta = self._run_single_evaluator(
                eval_func, name, outputs, inputs, ground_truth
            )
            metrics[name] = result
            if meta:
                metadata[name] = meta

        # Run async evaluators concurrently if any exist
        if async_evaluators:
            print('Evaluators cannot be run async. Please use sync evaluators only.')
        # if async_evaluators:

        #     async def arun_async_evaluators():
        #         async_tasks = [
        #             self._arun_single_evaluator(eval_func, name, outputs, inputs, ground_truth)
        #             for eval_func, name in async_evaluators
        #         ]
        #         return await asyncio.gather(*async_tasks)

        #     async_results = asyncio.run(arun_async_evaluators())

        #     for name, result, meta in async_results:
        #         metrics[name] = result
        #         if meta:
        #             metadata[name] = meta

        return metrics, metadata

    def _create_result(self, inputs, ground_truth, outputs, metrics, metadata):
        """Create standardized result dictionary."""
        return {
            'input': inputs,
            'ground_truth': ground_truth,
            'output': outputs,
            'metrics': metrics,
            'metadata': metadata,
        }

    def _get_inputs_and_ground_truth(self, datapoint_idx: int):
        """Get inputs and ground truth for evaluation from dataset."""
        if (
            self.use_hh_dataset
            and self.dataset.datapoints
            and len(self.dataset.datapoints) > 0
        ):
            datapoint_id = self.dataset.datapoints[datapoint_idx]
            datapoint_response = self.hhai.datapoints.get_datapoint(id=datapoint_id)
            return (
                datapoint_response.object.datapoint[0].inputs or {}, 
                datapoint_response.object.datapoint[0].ground_truth or {}
            )
        elif (
            self.dataset
            and len(self.dataset) > 0
        ):
            return (
                self.dataset[datapoint_idx].get('inputs', {}), 
                self.dataset[datapoint_idx].get('ground_truths', {})
            )
        return ({}, {})

    def _init_tracer(self, datapoint_idx: int, inputs: Dict[str, Any]) -> HoneyHiveTracer:
        """Initialize HoneyHiveTracer for evaluation."""
        hh = HoneyHiveTracer(
            api_key=self.api_key,
            project=self.project,
            source="evaluation",
            session_name=self.name,
            inputs={'inputs': inputs},
            is_evaluation=True,
            verbose=self.verbose,
            server_url=self.server_url,
            disable_http_tracing=self.disable_http_tracing,
            **self._get_tracing_metadata(datapoint_idx)
        )
        return hh

    def run_each(self, datapoint_idx: int) -> Dict[str, Any]:
        """Run evaluation for a single datapoint in its own thread."""

        inputs = {}
        ground_truth = {}
        outputs = None
        metrics = {}
        metadata = {}
        session_id = None

        # Get inputs
        try:
            inputs, ground_truth = self._get_inputs_and_ground_truth(datapoint_idx)
        except Exception as e:
            print(f"Error getting inputs for index {datapoint_idx}: {e}")
            return self._create_result(inputs, ground_truth, outputs, metrics, metadata)

        # Initialize tracer
        try:
            tracer = self._init_tracer(datapoint_idx, inputs)
            session_id = tracer.session_id
            self.evaluation_session_ids.append(session_id)
        except Exception as e:
            self.verbose and print(traceback.format_exc())
            raise Exception(
                f"Unable to initiate Honeyhive Tracer. Cannot run Evaluation: {e}"
            )

        # Run the function
        try:
            if inspect.iscoroutinefunction(self.function):
                raise ValueError("Evaluation task must be sync. Please use asyncio.run() to run coroutines inside the task.")
            else:
                outputs = self.function(inputs, ground_truth)
        except Exception as e:
            print(f"Error in evaluation function: {e}")
            print(traceback.format_exc())
        finally:
            HoneyHiveTracer.flush()
        
        # Run evaluators
        metrics, metadata = self._run_evaluators(outputs, inputs, ground_truth)
        
        # Add trace metadata, outputs, and metrics to session
        self._enrich_evaluation_session(
            datapoint_idx,
            session_id, 
            outputs,
            metrics,
            metadata
        )

        console.print(f"Test case {datapoint_idx} complete")
        
        return self._create_result(inputs, ground_truth, outputs, metrics, metadata)

    def run(self):
        """Public function to run the evaluation."""

        # create run
        eval_run = self.hhai.experiments.create_run(
            request=components.CreateRunRequest(
                project=self.project,
                name=self.name,
                dataset_id=self.dataset_id,
                event_ids=[],
                status=self.status,
                metadata=self.metadata
            )
        )
        self.eval_run = eval_run.create_run_response

        self.eval_result = EvaluationResult(
            run_id=self.eval_run.run_id,
            dataset_id=self.dataset_id,
            session_ids=[],
            status=self.status,
            suite=self.suite,
            stats={},
            data={},
        )

        #########################################################
        # Run evaluations
        #########################################################

        if self.use_hh_dataset:
            num_points = len(self.dataset.datapoints)
        else:
            num_points = len(self.dataset)

        
        start_time = time.time()
        if self.run_concurrently:
            with console.status("[bold green]Working on evals..."):
                # Use ThreadPoolExecutor to run evaluations concurrently
                with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                    try:
                        # Submit tasks and get futures with proper context propagation
                        futures = []
                        for i in range(num_points):
                            ctx = contextvars.copy_context()
                            futures.append(
                                executor.submit(
                                    ctx.run,
                                    functools.partial(self.run_each, i)
                                )
                            )
                        
                        # Collect results and log any errors
                        results = []
                        for future in futures:
                            try:
                                results.append(future.result())
                            except Exception as e:
                                print(f"Error in evaluation thread: {e}")
                                # Still add None result to maintain ordering
                                results.append(None)
                    except KeyboardInterrupt:
                        executor.shutdown(wait=False)
                        raise
                    finally:
                        HoneyHiveTracer.flush()

        else:
            results = []
            for i in range(num_points):
                result = self.run_each(i)
                results.append(result)

        end_time = time.time()
        #########################################################

        # Process results
        self.eval_result.stats = {
            'duration_s': round(end_time - start_time, 3),
        }
        self.eval_result.data = {
            'input': [],
            'output': [],
            'metrics': [],
            'metadata': [],
            'ground_truth': []
        }
        for r in results:
            for k in self.eval_result.data.keys():
                self.eval_result.data[k].append(r[k])

        # Convert deque to list after all threads complete
        self.eval_result.session_ids = list(self.evaluation_session_ids)

        #########################################################
        # Update run
        #########################################################
        try:
            if self.eval_run:
                self.status = "completed"
                
                # Only update with event_ids and status
                update_request = components.UpdateRunRequest(
                    event_ids=self.eval_result.session_ids, 
                    status=self.status
                )
                
                self.hhai.experiments.update_run(
                    run_id=self.eval_run.run_id,
                    update_run_request=update_request
                )
        except Exception:
            print("Warning: Unable to mark evaluation as `Completed`")

        if self.print_results:
            self.print_run()

    def print_run(self):
        """Print the results of the evaluation."""

        # get column names
        input_cols = {k for result in self.eval_result.data['input'] for k in result.keys()}
        metric_cols = {k for result in self.eval_result.data['metrics'] for k in result.keys()}
        metadata_cols = {k for result in self.eval_result.data['metadata'] for k in result.keys()}
        ground_truth_cols = {k for result in self.eval_result.data['ground_truth'] for k in result.keys()}

        # make table
        table = Table(
            title=f"Evaluation Results: {self.name}",
            show_lines=True,
            title_style=Style(
                color="black",
                bgcolor="yellow",
                bold=True,
                frame=True,
            ),
        )
        table.add_column("Suite", justify="center", style="magenta")
        for k in input_cols:
            table.add_column(f'Inputs.{k}', justify="center", style="green")
        table.add_column("Outputs", justify="center", style="blue")
        for k in ground_truth_cols:
            table.add_column(f'Ground Truths.{k}', justify="center", style="green")
        for k in metric_cols:
            table.add_column(f'Metrics.{k}', justify="center", style="blue")
        for k in metadata_cols:
            table.add_column(f'Metadata.{k}', justify="center", style="green")

        def truncated(string, max_length=500):
            if len(string) > max_length:
                return string[:max_length] + "..."
            return string

        # Get length of any list in data dict since they're all equal length
        n_rows = len(self.eval_result.data['input'])
        
        for idx in range(n_rows):
            row_values = [self.eval_result.suite]
            # Add input columns
            for k in input_cols:
                row_values.append(truncated(str(self.eval_result.data['input'][idx].get(k, ''))))
            # Add output column
            row_values.append(truncated(str(self.eval_result.data['output'][idx])))
            # Add ground truth columns
            for k in ground_truth_cols:
                row_values.append(truncated(str(self.eval_result.data['ground_truth'][idx].get(k, ''))))
            # Add metric columns
            for k in metric_cols:
                row_values.append(truncated(str(self.eval_result.data['metrics'][idx].get(k, ''))))
            # Add metadata columns
            for k in metadata_cols:
                row_values.append(truncated(str(self.eval_result.data['metadata'][idx].get(k, ''))))
            table.add_row(*row_values)

        console.print(table)

        # add footer with evaluation duration
        print(f"Evaluation Duration: {self.eval_result.stats['duration_s']} seconds\n")

        print('Exporting traces to HoneyHive...')


def evaluate(*args, **kwargs):

    eval = Evaluation(*args, **kwargs)

    # run evaluation
    eval.run()

    # print evaluation results
    if eval.print_results:
        eval.print_run()

    return EvaluationResult(
        run_id=eval.eval_run.run_id,
        dataset_id=eval.dataset_id,
        session_ids=eval.evaluation_session_ids,
        status=eval.status,
        data=eval.eval_result.data,
        stats=eval.eval_result.stats,
        suite=eval.suite
    )


__all__ = [
    "evaluate",
    "evaluator",
    "aevaluator",
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from enum import Enum
from honeyhive.types import BaseModel
from typing import Any, Dict, List, Optional, TypedDict
from typing_extensions import NotRequired


class MetricEditType(str, Enum):
    r"""Type of the metric - \"custom\", \"model\", \"human\" or \"composite\" """

    CUSTOM = "custom"
    MODEL = "model"
    HUMAN = "human"
    COMPOSITE = "composite"


class MetricEditReturnType(str, Enum):
    r"""The data type of the metric value - \"boolean\", \"float\", \"string\" """

    BOOLEAN = "boolean"
    FLOAT = "float"
    STRING = "string"


class MetricEditThresholdTypedDict(TypedDict):
    r"""Threshold for numeric metrics to decide passing or failing in tests"""

    min: NotRequired[float]
    max: NotRequired[float]


class MetricEditThreshold(BaseModel):
    r"""Threshold for numeric metrics to decide passing or failing in tests"""

    min: Optional[float] = None

    max: Optional[float] = None


class MetricEditEventType(str, Enum):
    r"""Type of event that the metric is set to be computed on"""

    MODEL = "model"
    TOOL = "tool"
    CHAIN = "chain"
    SESSION = "session"


class MetricEditTypedDict(TypedDict):
    metric_id: str
    r"""Unique identifier of the metric"""
    criteria: NotRequired[str]
    r"""Criteria for human or composite metrics"""
    name: NotRequired[str]
    r"""Updated name of the metric"""
    description: NotRequired[str]
    r"""Short description of what the metric does"""
    code_snippet: NotRequired[str]
    r"""Updated code block for the metric"""
    prompt: NotRequired[str]
    r"""Updated Evaluator prompt for the metric"""
    type: NotRequired[MetricEditType]
    r"""Type of the metric - \"custom\", \"model\", \"human\" or \"composite\" """
    enabled_in_prod: NotRequired[bool]
    r"""Whether to compute on all production events automatically"""
    needs_ground_truth: NotRequired[bool]
    r"""Whether a ground truth (on metadata) is required to compute it"""
    return_type: NotRequired[MetricEditReturnType]
    r"""The data type of the metric value - \"boolean\", \"float\", \"string\" """
    threshold: NotRequired[MetricEditThresholdTypedDict]
    r"""Threshold for numeric metrics to decide passing or failing in tests"""
    pass_when: NotRequired[bool]
    r"""Threshold for boolean metrics to decide passing or failing in tests"""
    event_name: NotRequired[str]
    r"""Name of event that the metric is set to be computed on"""
    event_type: NotRequired[MetricEditEventType]
    r"""Type of event that the metric is set to be computed on"""
    model_provider: NotRequired[str]
    r"""Provider of the model, formatted as a LiteLLM provider prefix"""
    model_name: NotRequired[str]
    r"""Name of the model, formatted as a LiteLLM model name"""
    child_metrics: NotRequired[List[Dict[str, Any]]]
    r"""Child metrics added under composite events"""


class MetricEdit(BaseModel):
    metric_id: str
    r"""Unique identifier of the metric"""

    criteria: Optional[str] = None
    r"""Criteria for human or composite metrics"""

    name: Optional[str] = None
    r"""Updated name of the metric"""

    description: Optional[str] = None
    r"""Short description of what the metric does"""

    code_snippet: Optional[str] = None
    r"""Updated code block for the metric"""

    prompt: Optional[str] = None
    r"""Updated Evaluator prompt for the metric"""

    type: Optional[MetricEditType] = None
    r"""Type of the metric - \"custom\", \"model\", \"human\" or \"composite\" """

    enabled_in_prod: Optional[bool] = None
    r"""Whether to compute on all production events automatically"""

    needs_ground_truth: Optional[bool] = None
    r"""Whether a ground truth (on metadata) is required to compute it"""

    return_type: Optional[MetricEditReturnType] = None
    r"""The data type of the metric value - \"boolean\", \"float\", \"string\" """

    threshold: Optional[MetricEditThreshold] = None
    r"""Threshold for numeric metrics to decide passing or failing in tests"""

    pass_when: Optional[bool] = None
    r"""Threshold for boolean metrics to decide passing or failing in tests"""

    event_name: Optional[str] = None
    r"""Name of event that the metric is set to be computed on"""

    event_type: Optional[MetricEditEventType] = None
    r"""Type of event that the metric is set to be computed on"""

    model_provider: Optional[str] = None
    r"""Provider of the model, formatted as a LiteLLM provider prefix"""

    model_name: Optional[str] = None
    r"""Name of the model, formatted as a LiteLLM model name"""

    child_metrics: Optional[List[Dict[str, Any]]] = None
    r"""Child metrics added under composite events"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from honeyhive.types import BaseModel
import pydantic
from typing import Any, Dict, List, Optional, TypedDict
from typing_extensions import Annotated, NotRequired


class DatapointTypedDict(TypedDict):
    id: NotRequired[str]
    r"""UUID for the datapoint"""
    tenant: NotRequired[str]
    project_id: NotRequired[str]
    r"""UUID for the project where the datapoint is stored"""
    created_at: NotRequired[str]
    updated_at: NotRequired[str]
    inputs: NotRequired[Dict[str, Any]]
    r"""Arbitrary JSON object containing the inputs for the datapoint"""
    history: NotRequired[List[Dict[str, Any]]]
    r"""Conversation history associated with the datapoint"""
    ground_truth: NotRequired[Dict[str, Any]]
    linked_event: NotRequired[str]
    r"""Event id for the event from which the datapoint was created"""
    linked_evals: NotRequired[List[str]]
    r"""Ids of evaluations where the datapoint is included"""
    linked_datasets: NotRequired[List[str]]
    r"""Ids of all datasets that include the datapoint"""
    saved: NotRequired[bool]
    type: NotRequired[str]
    r"""session or event - specify the type of data"""
    metadata: NotRequired[Dict[str, Any]]


class Datapoint(BaseModel):
    id: Annotated[Optional[str], pydantic.Field(alias="_id")] = None
    r"""UUID for the datapoint"""

    tenant: Optional[str] = None

    project_id: Optional[str] = None
    r"""UUID for the project where the datapoint is stored"""

    created_at: Optional[str] = None

    updated_at: Optional[str] = None

    inputs: Optional[Dict[str, Any]] = None
    r"""Arbitrary JSON object containing the inputs for the datapoint"""

    history: Optional[List[Dict[str, Any]]] = None
    r"""Conversation history associated with the datapoint"""

    ground_truth: Optional[Dict[str, Any]] = None

    linked_event: Optional[str] = None
    r"""Event id for the event from which the datapoint was created"""

    linked_evals: Optional[List[str]] = None
    r"""Ids of evaluations where the datapoint is included"""

    linked_datasets: Optional[List[str]] = None
    r"""Ids of all datasets that include the datapoint"""

    saved: Optional[bool] = None

    type: Optional[str] = None
    r"""session or event - specify the type of data"""

    metadata: Optional[Dict[str, Any]] = None



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from datetime import datetime
from honeyhive.types import BaseModel
import pydantic
from typing import List, Optional, TypedDict, Union
from typing_extensions import Annotated, NotRequired


OldValuesTypedDict = Union[float, bool]


OldValues = Union[float, bool]


NewValuesTypedDict = Union[float, bool]


NewValues = Union[float, bool]


class ExperimentComparisonResponseMetricsTypedDict(TypedDict):
    metric_name: NotRequired[str]
    event_name: NotRequired[str]
    metric_type: NotRequired[str]
    event_type: NotRequired[str]
    old_aggregate: NotRequired[float]
    new_aggregate: NotRequired[float]
    found_count: NotRequired[int]
    improved_count: NotRequired[int]
    degraded_count: NotRequired[int]
    same_count: NotRequired[int]
    improved: NotRequired[List[str]]
    degraded: NotRequired[List[str]]
    same: NotRequired[List[str]]
    old_values: NotRequired[List[OldValuesTypedDict]]
    new_values: NotRequired[List[NewValuesTypedDict]]


class ExperimentComparisonResponseMetrics(BaseModel):
    metric_name: Optional[str] = None

    event_name: Optional[str] = None

    metric_type: Optional[str] = None

    event_type: Optional[str] = None

    old_aggregate: Optional[float] = None

    new_aggregate: Optional[float] = None

    found_count: Optional[int] = None

    improved_count: Optional[int] = None

    degraded_count: Optional[int] = None

    same_count: Optional[int] = None

    improved: Optional[List[str]] = None

    degraded: Optional[List[str]] = None

    same: Optional[List[str]] = None

    old_values: Optional[List[OldValues]] = None

    new_values: Optional[List[NewValues]] = None


class EventDetailsTypedDict(TypedDict):
    event_name: NotRequired[str]
    event_type: NotRequired[str]
    presence: NotRequired[str]


class EventDetails(BaseModel):
    event_name: Optional[str] = None

    event_type: Optional[str] = None

    presence: Optional[str] = None


class EvaluatorsTypedDict(TypedDict):
    pass


class Evaluators(BaseModel):
    pass


class ExperimentComparisonResponseResultsTypedDict(TypedDict):
    pass


class ExperimentComparisonResponseResults(BaseModel):
    pass


class ExperimentComparisonResponseSchemasConfigurationTypedDict(TypedDict):
    pass


class ExperimentComparisonResponseSchemasConfiguration(BaseModel):
    pass


class MetadataTypedDict(TypedDict):
    pass


class Metadata(BaseModel):
    pass


class PassingRangesTypedDict(TypedDict):
    pass


class PassingRanges(BaseModel):
    pass


class OldRunTypedDict(TypedDict):
    id: NotRequired[str]
    run_id: NotRequired[str]
    project: NotRequired[str]
    tenant: NotRequired[str]
    created_at: NotRequired[datetime]
    event_ids: NotRequired[List[str]]
    session_ids: NotRequired[List[str]]
    dataset_id: NotRequired[str]
    datapoint_ids: NotRequired[List[str]]
    evaluators: NotRequired[List[EvaluatorsTypedDict]]
    results: NotRequired[ExperimentComparisonResponseResultsTypedDict]
    configuration: NotRequired[
        ExperimentComparisonResponseSchemasConfigurationTypedDict
    ]
    metadata: NotRequired[MetadataTypedDict]
    passing_ranges: NotRequired[PassingRangesTypedDict]
    status: NotRequired[str]
    name: NotRequired[str]


class OldRun(BaseModel):
    id: Annotated[Optional[str], pydantic.Field(alias="_id")] = None

    run_id: Optional[str] = None

    project: Optional[str] = None

    tenant: Optional[str] = None

    created_at: Optional[datetime] = None

    event_ids: Optional[List[str]] = None

    session_ids: Optional[List[str]] = None

    dataset_id: Optional[str] = None

    datapoint_ids: Optional[List[str]] = None

    evaluators: Optional[List[Evaluators]] = None

    results: Optional[ExperimentComparisonResponseResults] = None

    configuration: Optional[ExperimentComparisonResponseSchemasConfiguration] = None

    metadata: Optional[Metadata] = None

    passing_ranges: Optional[PassingRanges] = None

    status: Optional[str] = None

    name: Optional[str] = None


class ExperimentComparisonResponseEvaluatorsTypedDict(TypedDict):
    pass


class ExperimentComparisonResponseEvaluators(BaseModel):
    pass


class ExperimentComparisonResponseSchemasResultsTypedDict(TypedDict):
    pass


class ExperimentComparisonResponseSchemasResults(BaseModel):
    pass


class ExperimentComparisonResponseConfigurationTypedDict(TypedDict):
    pass


class ExperimentComparisonResponseConfiguration(BaseModel):
    pass


class ExperimentComparisonResponseMetadataTypedDict(TypedDict):
    pass


class ExperimentComparisonResponseMetadata(BaseModel):
    pass


class ExperimentComparisonResponsePassingRangesTypedDict(TypedDict):
    pass


class ExperimentComparisonResponsePassingRanges(BaseModel):
    pass


class NewRunTypedDict(TypedDict):
    id: NotRequired[str]
    run_id: NotRequired[str]
    project: NotRequired[str]
    tenant: NotRequired[str]
    created_at: NotRequired[datetime]
    event_ids: NotRequired[List[str]]
    session_ids: NotRequired[List[str]]
    dataset_id: NotRequired[str]
    datapoint_ids: NotRequired[List[str]]
    evaluators: NotRequired[List[ExperimentComparisonResponseEvaluatorsTypedDict]]
    results: NotRequired[ExperimentComparisonResponseSchemasResultsTypedDict]
    configuration: NotRequired[ExperimentComparisonResponseConfigurationTypedDict]
    metadata: NotRequired[ExperimentComparisonResponseMetadataTypedDict]
    passing_ranges: NotRequired[ExperimentComparisonResponsePassingRangesTypedDict]
    status: NotRequired[str]
    name: NotRequired[str]


class NewRun(BaseModel):
    id: Annotated[Optional[str], pydantic.Field(alias="_id")] = None

    run_id: Optional[str] = None

    project: Optional[str] = None

    tenant: Optional[str] = None

    created_at: Optional[datetime] = None

    event_ids: Optional[List[str]] = None

    session_ids: Optional[List[str]] = None

    dataset_id: Optional[str] = None

    datapoint_ids: Optional[List[str]] = None

    evaluators: Optional[List[ExperimentComparisonResponseEvaluators]] = None

    results: Optional[ExperimentComparisonResponseSchemasResults] = None

    configuration: Optional[ExperimentComparisonResponseConfiguration] = None

    metadata: Optional[ExperimentComparisonResponseMetadata] = None

    passing_ranges: Optional[ExperimentComparisonResponsePassingRanges] = None

    status: Optional[str] = None

    name: Optional[str] = None


class ExperimentComparisonResponseTypedDict(TypedDict):
    metrics: NotRequired[List[ExperimentComparisonResponseMetricsTypedDict]]
    common_datapoints: NotRequired[List[str]]
    event_details: NotRequired[List[EventDetailsTypedDict]]
    old_run: NotRequired[OldRunTypedDict]
    new_run: NotRequired[NewRunTypedDict]


class ExperimentComparisonResponse(BaseModel):
    metrics: Optional[List[ExperimentComparisonResponseMetrics]] = None

    common_datapoints: Annotated[
        Optional[List[str]], pydantic.Field(alias="commonDatapoints")
    ] = None

    event_details: Optional[List[EventDetails]] = None

    old_run: Optional[OldRun] = None

    new_run: Optional[NewRun] = None



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from honeyhive.models.components import (
    createrunresponse as components_createrunresponse,
)
from honeyhive.types import BaseModel
import httpx
from typing import Optional, TypedDict
from typing_extensions import NotRequired


class CreateRunResponseTypedDict(TypedDict):
    content_type: str
    r"""HTTP response content type for this operation"""
    status_code: int
    r"""HTTP response status code for this operation"""
    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""
    create_run_response: NotRequired[
        components_createrunresponse.CreateRunResponseTypedDict
    ]
    r"""Successful response"""


class CreateRunResponse(BaseModel):
    content_type: str
    r"""HTTP response content type for this operation"""

    status_code: int
    r"""HTTP response status code for this operation"""

    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""

    create_run_response: Optional[components_createrunresponse.CreateRunResponse] = None
    r"""Successful response"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from enum import Enum
from honeyhive.models.components import (
    experimentresultresponse as components_experimentresultresponse,
)
from honeyhive.types import BaseModel
from honeyhive.utils import FieldMetadata, PathParamMetadata, QueryParamMetadata
import httpx
from typing import Optional, TypedDict
from typing_extensions import Annotated, NotRequired


class AggregateFunction(str, Enum):
    AVERAGE = "average"
    MIN = "min"
    MAX = "max"
    MEDIAN = "median"
    P95 = "p95"
    P99 = "p99"
    P90 = "p90"
    SUM = "sum"
    COUNT = "count"


class GetExperimentResultRequestTypedDict(TypedDict):
    run_id: str
    project_id: str
    aggregate_function: NotRequired[AggregateFunction]


class GetExperimentResultRequest(BaseModel):
    run_id: Annotated[
        str, FieldMetadata(path=PathParamMetadata(style="simple", explode=False))
    ]

    project_id: Annotated[
        str, FieldMetadata(query=QueryParamMetadata(style="form", explode=True))
    ]

    aggregate_function: Annotated[
        Optional[AggregateFunction],
        FieldMetadata(query=QueryParamMetadata(style="form", explode=True)),
    ] = None


class GetExperimentResultResponseTypedDict(TypedDict):
    content_type: str
    r"""HTTP response content type for this operation"""
    status_code: int
    r"""HTTP response status code for this operation"""
    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""
    experiment_result_response: NotRequired[
        components_experimentresultresponse.ExperimentResultResponseTypedDict
    ]
    r"""Experiment result retrieved successfully"""


class GetExperimentResultResponse(BaseModel):
    content_type: str
    r"""HTTP response content type for this operation"""

    status_code: int
    r"""HTTP response status code for this operation"""

    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""

    experiment_result_response: Optional[
        components_experimentresultresponse.ExperimentResultResponse
    ] = None
    r"""Experiment result retrieved successfully"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from honeyhive.models.components import (
    createeventrequest as components_createeventrequest,
)
from honeyhive.types import BaseModel
import httpx
from typing import Optional, TypedDict
from typing_extensions import NotRequired


class CreateEventRequestBodyTypedDict(TypedDict):
    event: NotRequired[components_createeventrequest.CreateEventRequestTypedDict]


class CreateEventRequestBody(BaseModel):
    event: Optional[components_createeventrequest.CreateEventRequest] = None


class CreateEventResponseBodyTypedDict(TypedDict):
    r"""Event created"""

    event_id: NotRequired[str]
    success: NotRequired[bool]


class CreateEventResponseBody(BaseModel):
    r"""Event created"""

    event_id: Optional[str] = None

    success: Optional[bool] = None


class CreateEventResponseTypedDict(TypedDict):
    content_type: str
    r"""HTTP response content type for this operation"""
    status_code: int
    r"""HTTP response status code for this operation"""
    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""
    object: NotRequired[CreateEventResponseBodyTypedDict]
    r"""Event created"""


class CreateEventResponse(BaseModel):
    content_type: str
    r"""HTTP response content type for this operation"""

    status_code: int
    r"""HTTP response status code for this operation"""

    raw_response: httpx.Response
    r"""Raw HTTP response; suitable for custom response parsing"""

    object: Optional[CreateEventResponseBody] = None
    r"""Event created"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from honeyhive.types import BaseModel
from typing import Any, Dict, List, Optional, TypedDict
from typing_extensions import NotRequired


class CreateModelEventTypedDict(TypedDict):
    project: str
    r"""Project associated with the event"""
    model: str
    r"""Model name"""
    provider: str
    r"""Model provider"""
    messages: List[Dict[str, Any]]
    r"""Messages passed to the model"""
    response: Dict[str, Any]
    r"""Final output JSON of the event"""
    duration: float
    r"""How long the event took in milliseconds"""
    usage: Dict[str, Any]
    r"""Usage statistics of the model"""
    cost: NotRequired[float]
    r"""Cost of the model completion"""
    error: NotRequired[str]
    r"""Any error description if event failed"""
    source: NotRequired[str]
    r"""Source of the event - production, staging, etc"""
    event_name: NotRequired[str]
    r"""Name of the event"""
    hyperparameters: NotRequired[Dict[str, Any]]
    r"""Hyperparameters used for the model"""
    template: NotRequired[List[Dict[str, Any]]]
    r"""Template used for the model"""
    template_inputs: NotRequired[Dict[str, Any]]
    r"""Inputs for the template"""
    tools: NotRequired[List[Dict[str, Any]]]
    r"""Tools used for the model"""
    tool_choice: NotRequired[str]
    r"""Tool choice for the model"""
    response_format: NotRequired[Dict[str, Any]]
    r"""Response format for the model"""


class CreateModelEvent(BaseModel):
    project: str
    r"""Project associated with the event"""

    model: str
    r"""Model name"""

    provider: str
    r"""Model provider"""

    messages: List[Dict[str, Any]]
    r"""Messages passed to the model"""

    response: Dict[str, Any]
    r"""Final output JSON of the event"""

    duration: float
    r"""How long the event took in milliseconds"""

    usage: Dict[str, Any]
    r"""Usage statistics of the model"""

    cost: Optional[float] = None
    r"""Cost of the model completion"""

    error: Optional[str] = None
    r"""Any error description if event failed"""

    source: Optional[str] = None
    r"""Source of the event - production, staging, etc"""

    event_name: Optional[str] = None
    r"""Name of the event"""

    hyperparameters: Optional[Dict[str, Any]] = None
    r"""Hyperparameters used for the model"""

    template: Optional[List[Dict[str, Any]]] = None
    r"""Template used for the model"""

    template_inputs: Optional[Dict[str, Any]] = None
    r"""Inputs for the template"""

    tools: Optional[List[Dict[str, Any]]] = None
    r"""Tools used for the model"""

    tool_choice: Optional[str] = None
    r"""Tool choice for the model"""

    response_format: Optional[Dict[str, Any]] = None
    r"""Response format for the model"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from typing import Any

def get_discriminator(model: Any, fieldname: str, key: str) -> str:
    if isinstance(model, dict):
        try:
            return f'{model.get(key)}'
        except AttributeError as e:
            raise ValueError(f'Could not find discriminator key {key} in {model}') from e

    if hasattr(model, fieldname):
        return f'{getattr(model, fieldname)}'

    fieldname = fieldname.upper()
    if hasattr(model, fieldname):
        return f'{getattr(model, fieldname)}'

    raise ValueError(f'Could not find discriminator field {fieldname} in {model}')
